"""
NodePy (Numerical ODE solvers in Python) is...
"""

from __future__ import absolute_import
__version__="1.0.0"

import nodepy.runge_kutta_method as rk
import nodepy.twostep_runge_kutta_method as tsrk
import nodepy.downwind_runge_kutta_method as dwrk
import nodepy.linear_multistep_method as lm
import nodepy.rooted_trees as rt
import nodepy.ivp
import nodepy.convergence as conv
import nodepy.low_storage_rk as lsrk
import nodepy.graph
import nodepy.loadmethod



from __future__ import absolute_import
from six.moves import range
def bisect(rlo, rhi, acc, tol, fun, **kwargs):
    """ 
        Performs a bisection search.

        **Input**:
            - fun -- a function such that fun(r)==True iff x_0>r, where x_0 is the value to be found.
    """
    while rhi-rlo>acc:
        r=0.5*(rhi+rlo)
        if kwargs: isvalid=fun(r,tol,**kwargs)
        else: isvalid=fun(r,tol)
        if isvalid:
            rlo=r
        else:
            rhi=r
    return rlo

def permutations(str):
    if len(str) <=1:
        yield str
    else:
        for perm in permutations(str[1:]):
            for i in range(len(perm)+1):
                yield perm[:i] + str[0:1] + perm[i:]

def shortstring(x,printzeros=False):
    import numpy as np
    import sympy.core.numbers
    if x==0 and printzeros==False:
        return ''
    elif x.__class__ is np.float64 or x.__class__ is sympy.Float:
        return '%6.3f' % x
    else: 
        return ' '+str(x)

def array2strings(x,printzeros=False):
    import numpy as np
    return np.reshape([shortstring(xi,printzeros) for xi in x.reshape(-1)],x.shape)

def find_plot_bounds(f,guess,N=101,zmax=1000):
    r"""Find reasonable area to plot for stability regions.
    
    Tries to find an area that contains the entire region where f returns True,
    but isn't too big.  Makes lots of assumptions and is really intended for
    finding stability regions.  Obviously can't work for unbounded stability
    regions.

    f should return True if f(z) is in the stability region.

    N should be odd in order to catch very small stability regions.
    """
    import numpy as np

    bounds = guess
    old_bounds = []

    while bounds != old_bounds:
        old_bounds = bounds
        y=np.linspace(bounds[2],bounds[3],N)

        #Check boundaries
        bounds = list(bounds)
        close = False
        while abs(bounds[0])<zmax:
            x=np.linspace(bounds[0],bounds[1],N)
            Z=x[0]+y*1j

            left = f(Z)
            if np.any(left):
                bounds[0] = 1.5*bounds[0]
                close = True
            else:
                if close == True:
                    break
                bounds[0] = bounds[0]/1.5
                if bounds[0] > -1.e-15:
                    return bounds

        bounds[1] = -0.1*bounds[0]
        x=np.linspace(bounds[0],bounds[1],N) + 0.*1j

        close = False
        while abs(bounds[2])<zmax:
            y=np.linspace(bounds[2],bounds[3],N)
            Z=x + y[0]*1j

            bottom = f(Z)
            if np.any(bottom):
                bounds[2] = 1.5*bounds[2]
                close = True
            else:
                if close == True:
                    break
                bounds[2] = bounds[2]/1.5
                if bounds[2] > -1.e-15:
                    return bounds

        bounds[3] = -bounds[2]

    return bounds

 



r"""
Class for two-step Runge-Kutta methods, and various functions related to them.

This module is extremely experimental and some parts may be incompatible
with the rest of nodepy.

**Examples**::

    >>> from nodepy import twostep_runge_kutta_method as tsrk

* Load methods::

    >>> tsrk4 = tsrk.loadTSRK('order4')
    >>> tsrk5 = tsrk.loadTSRK('order5')

    >>> print(tsrk4)
    Two-step Runge-Kutta Method
    General
     -113/88      | 1435/352      -479/352      | 1
     -103/88      | 1917/352      -217/352      |               1
    ______________|_____________________________|_____________________________
    -4483/8011    | 180991/96132  -17777/32044  | -44709/32044  48803/96132
    >>> print(tsrk4.latex())
    \begin{align}
      \begin{array}{c|cc|cc}
      - \frac{113}{88} & \frac{1435}{352} & - \frac{479}{352} & 1 & 0\\
      - \frac{103}{88} & \frac{1917}{352} & - \frac{217}{352} & 0 & 1\\
      \hline
      - \frac{4483}{8011} & \frac{180991}{96132} & - \frac{17777}{32044} & - \frac{44709}{32044} & \frac{48803}{96132}
      \end{array}
    \end{align}

    >>> tsrk4.plot_stability_region()

* Check their order of accuracy::

    >>> tsrk4.order()
    4
    >>> tsrk5.order(tol=1.e-3)
    5

* Get the radius of absolute monotonicity::

    >>> tsrk4.absolute_monotonicity_radius()
    0
    >>> tsrk5.absolute_monotonicity_radius()
    0


**References**:
    [jackiewicz1995,butcher1997,hairer1997]

"""
from __future__ import print_function
from __future__ import division

from __future__ import absolute_import
import numpy as np

import nodepy.rooted_trees as rt
import nodepy.snp as snp
from nodepy.strmanip import *
from nodepy.general_linear_method import GeneralLinearMethod
from six.moves import range

#=====================================================
class TwoStepRungeKuttaMethod(GeneralLinearMethod):
#=====================================================
    r""" General class for Two-step Runge-Kutta Methods
        The representation uses the form and partly the notation of :cite:`jackiewicz1995`,
        equation (1.3).

        `\begin{align*}
        y^n_j = & d_j u^{n-1} + (1-d_j)u^n + \Delta t \sum_{k=1}^{s}
        (\hat{a}_{jk} f(y_k^{n-1}) + a_{jk} f(y_k^n)) & (1\le j \le s) \\
        u^{n+1} = & \theta u^{n-1} + (1-\theta)u^n + \Delta t \sum_{j=1}^{s}(\hat{b}_j f(y_j^{n-1}) + b_j f(y_j^n))
        \end{align*}`
    """
    def __init__(self,d,theta,A,b,Ahat=None,bhat=None,type='General',name='Two-step Runge-Kutta Method'):
        r"""
            Initialize a 2-step Runge-Kutta method."""
        d,A,b,Ahat,bhat=snp.normalize(d,A,b,Ahat,bhat)
        self.s = max(np.shape(b))
        self.d,self.theta,self.A,self.b = d,theta,A,b
        self.Ahat,self.bhat=Ahat,bhat
        #if type=='General': self.Ahat,self.bhat=Ahat,bhat
        #elif type=='Type I':
        #    self.Ahat = np.zeros([self.s,self.s])
        #    self.bhat = np.zeros([self.s,1])
        #elif type=='Type II':
        #    self.Ahat = np.zeros([self.s,self.s])
        #    self.Ahat[:,0]=Ahat
        #    self.bhat = np.zeros([self.s,1])
        #    self.bhat[0]=bhat
        #else: raise Exception('Unrecognized type')
        self.name=name
        self.type=type

    def order(self,tol=1.e-13):
        r"""
            Return the order of a Two-step Runge-Kutta method.
            Computed by computing the elementary weights corresponding
            to the appropriate rooted trees.
        """
        p=0
        while True:
            z=self.order_conditions(p+1)
            if np.any(abs(z)>tol): return p
            p=p+1

    def __len__(self):
        """
            The length of the method is the number of stages.
        """
        return np.size(self.A,0)

    def order_conditions(self,p):
        r"""
            Evaluate the order conditions corresponding to rooted trees
            of order $p$.

            **Output**:
                - A vector $z$ of residuals (the amount by which each
                  order condition is violated)

            TODO: Implement simple order conditions (a la Albrecht)
                    for Type I & II TSRKs
        """
        from numpy import dot
        d,theta,Ahat,A,bhat,b=self.d,self.theta,self.Ahat,self.A,self.bhat,self.b
        e=np.ones(len(b))
        b=b.T; bhat=bhat.T
        c=dot(Ahat+A,e)-d
        code=TSRKOrderConditions(p)
        z=np.zeros(len(code))
        for i in range(len(code)):
            exec('yy='+code[i])
            exec('z[i]='+code[i])
        return z


    def stability_matrix(self,z):
        r"""
            Constructs the stability matrix of a two-step Runge-Kutta method.
            Right now just for a specific value of z.
            We ought to use Sage to do it symbolically.

            **Output**:
                M -- stability matrix evaluated at z

            WARNING: This only works for Type I & Type II methods
            right now!!!
        """
        s = self.Ahat.shape[1]
        if self.type == 'General':
            # J Y^n = K Y^{n-1}
            K1 = np.column_stack((z*self.Ahat,self.d,1-self.d))
            K2 = snp.zeros(s+2); K2[-1] = 1
            K3 = np.concatenate((z*self.bhat,np.array((self.theta,1-self.theta))))
            K = np.vstack((K1,K2,K3))

            J = snp.eye(s+2)
            J[:s,:s] = J[:s,:s] - z*self.A
            J[-1,:s] = z*self.b

            M = snp.solve(J.astype('complex64'),K.astype('complex64'))
            #M = snp.solve(J, K) # This version is slower

        else:
            D=np.hstack([1.-self.d,self.d])
            thet=np.hstack([1.-self.theta,self.theta])
            A,b=self.A,self.b
            if self.type=='Type II':
                ahat = np.zeros([self.s,1]); ahat[:,0] = self.Ahat[:,0]
                bh = np.zeros([1,1]); bh[0,0]=self.bhat[0]
                A = np.hstack([ahat,self.A])
                A = np.vstack([np.zeros([1,self.s+1]),A])
                b =  np.vstack([bh,self.b])

            M1=np.linalg.solve(np.eye(self.s)-z*self.A,D)
            L1=thet+z*np.dot(self.b.T,M1)
            M=np.vstack([L1,[1.,0.]])
        return M

    def __num__(self):
        """
        Returns a copy of the method but with floating-point coefficients.
        This is useful whenever we need to operate numerically without
        worrying about the representation of the method.
        """
        import copy
        numself = copy.deepcopy(self)
        if self.A.dtype==object:
            for coeff_array in ['A','Ahat','b','bhat','theta','d']:
                setattr(numself,coeff_array,np.array(getattr(self,coeff_array),dtype=np.float64))
        return numself

    def plot_stability_region(self,N=50,bounds=None,
                    color='r',filled=True,scaled=False):
        r"""
            Plot the region of absolute stability
            of a Two-step Runge-Kutta method, i.e. the set

            `\{ z \in C : M(z) is power bounded \}`

            where $M(z)$ is the stability matrix of the method.

            **Input**: (all optional)
                - N       -- Number of gridpoints to use in each direction
                - bounds  -- limits of plotting region
                - color   -- color to use for this plot
                - filled  -- if true, stability region is filled in (solid); otherwise it is outlined
        """
        method = self.__num__() # Use floating-point coefficients for efficiency

        import matplotlib.pyplot as plt
        if bounds is None:
            from nodepy.utils import find_plot_bounds
            stable = lambda z : max(abs(np.linalg.eigvals(method.stability_matrix(z))))<=1.0
            bounds = find_plot_bounds(np.vectorize(stable),guess=(-10,1,-5,5))
            if np.min(np.abs(np.array(bounds)))<1.e-14:
                print('No stable region found; is this method zero-stable?')

        x=np.linspace(bounds[0],bounds[1],N)
        y=np.linspace(bounds[2],bounds[3],N)
        X=np.tile(x,(N,1))
        Y=np.tile(y[:,np.newaxis],(1,N))
        Z=X+Y*1j

        maxroot = lambda z : max(abs(np.linalg.eigvals(method.stability_matrix(z))))
        Mroot = np.vectorize(maxroot)
        R = Mroot(Z)

        if filled:
            plt.contourf(X,Y,R,[0,1],colors=color)
        else:
            plt.contour(X,Y,R,[0,1],colors=color)
        plt.title('Absolute Stability Region for '+self.name)
        plt.plot([0,0],[bounds[2],bounds[3]],'--k',linewidth=2)
        plt.plot([bounds[0],bounds[1]],[0,0],'--k',linewidth=2)
        plt.axis('Image')

    def absolute_monotonicity_radius(self,acc=1.e-10,rmax=200,
                    tol=3.e-16):
        r"""
            Returns the radius of absolute monotonicity
            of a TSRK method.
        """
        from nodepy.utils import bisect
        r=bisect(0,rmax,acc,tol,self.is_absolutely_monotonic)
        return r

    def latex(self):
        """A laTeX representation of the compact form."""
        from sympy.printing import latex

        d       = self.d
        A       = self.A
        Ahat    = self.Ahat
        b       = self.b
        bhat    = self.bhat
        theta   = self.theta

        s= r'\begin{align}'
        s+='\n'
        s+=r'  \begin{array}{c|'
        s+='c'*(len(self)) +'|'
        s+='c'*(len(self))
        s+='}\n'
        for i in range(len(self)):
            s+='  '+latex(d[i])

            for j in range(len(self)):
                s+=' & '+latex(Ahat[i,j])

            for j in range(len(self)):
                s+=' & '+latex(A[i,j])

            s+=r'\\'
            s+='\n'
        s+=r'  \hline'
        s+='\n'
        s+= '  '+latex(theta)
        for j in range(len(self)):
            s+=' & '+latex(bhat[j])
        for j in range(len(self)):
            s+=' & '+latex(b[j])
        s+='\n'
        s+=r'  \end{array}'
        s+='\n'
        s+=r'\end{align}'
        s=s.replace('- -','')
        return s


    def __str__(self):
        from nodepy.utils import array2strings
        from nodepy.runge_kutta_method import _get_column_widths

        d       = array2strings(self.d)
        A       = array2strings(self.A)
        Ahat    = array2strings(self.Ahat)
        b       = array2strings(self.b)
        bhat    = array2strings(self.bhat)

        theta = str(self.theta)
        lenmax, colmax = _get_column_widths([d, Ahat, A, bhat, b])

        s   = self.name+'\n'+self.type+'\n'
        for i in range(len(self)):
                s+=d[i].ljust(colmax+1)+'|'
                for j in range(len(self)):
                        s+=Ahat[i,j].ljust(colmax+1)
                s+=' |'
                for j in range(len(self)):
                        s+=A[i,j].ljust(colmax+1)
                s=s.rstrip()+'\n'
        s+='_'*(colmax+1)+('|_'+'_'*(colmax+1)*np.size(A,0))*2+'\n'

        s+= theta.ljust(colmax)
        s+=' |'
        for j in range(len(self)):
                s+=bhat[j].ljust(colmax+1)
        s+=' |'
        for j in range(len(self)):
                s+=b[j].ljust(colmax+1)
        return s.rstrip()

    def spijker_form(self):
        r""" Returns arrays $S,T$ such that the TSRK can be written
            $$ w = S x + T f(w),$$
            and such that $\[S \ \ T\]$ has no two rows equal.
            See the TSRK paper by Ketcheson, Gottlieb, and Macdonald.
            Equation (2.5) therein gives the Spijker form for general TSRKs,
            while the last (unnumbered) equation of Section 4.2 gives the
            form for TSRKs of Type II (which are the type described by (4.2)).
        """
        s=self.s
        if self.type=='General':
            zero_column = np.zeros( (s,1) )
            T3 = np.hstack( (self.Ahat,zero_column,self.A,zero_column) )
            T4 = np.hstack( (self.bhat, 0, self.b, 0) )
            T = np.vstack( (np.zeros([s+1,2*s+2]),T3,T4) )

            S1 = np.hstack( (np.zeros([s+1,1]),np.eye(s+1)) )
            S2 = np.column_stack( (self.d,np.zeros([s,s]),1-self.d) )
            S3 = np.hstack( ([[self.theta]],np.zeros([1,s]),[[1-self.theta]]) )
            S = np.vstack( (S1,S2,S3) )

        elif self.type=='Type I':
            K = np.vstack([self.A,self.b.T])
            T2 = np.hstack([np.zeros([s+1,1]),K,np.zeros([s+1,1])])
            T =  np.vstack([np.zeros([1,s+1]),T2])
            S1 = np.vstack([np.zeros([1,1]),self.d,np.array([[self.theta]])])
            S2 = np.vstack([np.zeros([1,1]),1-self.d,np.array([[1-self.theta]])])
            S = np.hstack([S1,S2])

        elif self.type=='Type II':
            S0 = np.array([1,0])
            S1 = np.hstack([self.d,1-self.d])
            S2 = np.array([self.theta,1-self.theta])
            S =  np.vstack([S0,S1,S2])
            ahat = np.zeros([s,1])
            ahat[:,0] = self.Ahat[:,0]
            bh = np.zeros([1,1])
            bh[0,0]=self.bhat[0]
            T0 = np.zeros([1,s+2])
            T1 =  np.hstack([ahat,self.A,np.zeros([s,1])])
            T2 =  np.hstack([bh,self.b.T,np.zeros([1,1])])
            T = np.vstack([T0,T1,T2])

        return S.astype(np.float64), T.astype(np.float64)


    def is_absolutely_monotonic(self,r,tol):
        r""" Returns 1 if the TSRK method is absolutely monotonic
            at $z=-r$.

            The method is absolutely monotonic if $(I+rT)^{-1}$ exists
            and
            $$(I+rT)^{-1}T \\ge 0$$
            $$(I+rT)^{-1}S \\ge 0$$

            The inequalities are interpreted componentwise.

            **References**:
                #. [spijker2007]
        """
        S, T = self.spijker_form()
        m = np.shape(T)[0]
        X = np.eye(m) + r * T
        if abs(np.linalg.det(X)) < tol:
            return 0
        P = np.linalg.solve(X, T)
        R = np.linalg.solve(X, S)
        if P.min()<-tol or R.min()<-tol:
            return 0
        else:
            return 1
        # Need an exception here if rhi==rmax


#================================================================
# Functions for analyzing Two-step Runge-Kutta order conditions
#================================================================

def TSRKOrderConditions(p,ind='all'):
    from nodepy.rooted_trees import Emap_str
    forest=rt.list_trees(p)
    code=[]
    for tree in forest:
        code.append(tsrk_elementary_weight_str(tree)+'-'+Emap_str(tree))
        code[-1]=code[-1].replace('--','')
        code[-1]=code[-1].replace('1 ','e ')
        code[-1]=code[-1].replace('1)','e)')
    return code

def tsrk_elementary_weight(tree):
    """
        Constructs Butcher's elementary weights
        for Two-step Runge-Kutta methods
    """
    from sympy import Symbol
    bhat,b,theta=Symbol('bhat',False),Symbol('b',False),Symbol('theta',False)
    ew=bhat*tree.Gprod(rt.Emap,rt.Gprod,betaargs=[TSRKeta,Dmap],alphaargs=[-1])+b*tree.Gprod(TSRKeta,rt.Dmap)+theta*tree.Emap(-1)
    return ew

def tsrk_elementary_weight_str(tree):
    """
        Constructs Butcher's elementary weights
        for Two-step Runge-Kutta methods
        as numpy-executable strings
    """
    from nodepy.rooted_trees import Dmap_str, Emap_str
    ewstr='dot(bhat,'+tree.Gprod_str(rt.Emap_str,rt.Gprod_str,betaargs=[TSRKeta_str,Dmap_str],alphaargs=[-1])+')+dot(b,'+tree.Gprod_str(TSRKeta_str,rt.Dmap_str)+')+theta*'+Emap_str(tree,-1)
    ewstr=mysimp(ewstr)
    return ewstr

def TSRKeta(tree):
    from nodepy.rooted_trees import Dprod
    from sympy import symbols
    raise Exception('This function does not work correctly; use the _str version')
    if tree=='':  return 1
    if tree=='T': return symbols('c',commutative=False)
    return symbols('d',commutative=False)*tree.Emap(-1)+symbols('Ahat',commutative=False)*tree.Gprod(Emap,Dprod,betaargs=[TSRKeta],alphaargs=[-1])+symbols('A',commutative=False)*Dprod(tree,TSRKeta)

def TSRKeta_str(tree):
    """
    Computes eta(t) for Two-step Runge-Kutta methods
    """
    from nodepy.rooted_trees import Dprod_str, Emap_str
    if tree=='':  return 'e'
    if tree=='T': return 'c'
    return '(d*'+Emap_str(tree,-1)+'+dot(Ahat,'+tree.Gprod_str(Emap_str,Dprod_str,betaargs=[TSRKeta_str],alphaargs=[-1])+')'+'+dot(A,'+Dprod_str(tree,TSRKeta_str)+'))'


#================================================================

def loadTSRK(which='All'):
    r"""
        Load two particular TSRK methods (From :cite:`jackiewicz1995`).

        The method of order five satisfies the order conditions only
        to four or five digits of accuracy.
    """
    from sympy import Rational
    one  = Rational(1,1)

    TSRK={}
    #================================================
    c = one
    lamda = 0
    theta = (6*c**2 - 12*c + 5)/(1 - 6*c**2)
    d    = np.array( [ (2*c - c**2 -2*lamda)/(2*c - 1) ] )
    Ahat = np.array( [[(c+c**2-lamda-2*c*lamda)/(2*c - 1)]] )
    A    = np.array( [[lamda]] )
    bhat = np.array( [2*(1-3*c**2)/(1-6*c**2)] )
    b    = np.array( [2*(2-6*c+3*c**2)/(1-6*c**2)] )
    TSRK['order3']=TwoStepRungeKuttaMethod(d,theta,A,b,Ahat,bhat,type='General')
    #================================================
    d=np.array([-113*one/88,-103*one/88])
    theta=-4483*one/8011
    Ahat=np.array([[1435*one/352,-479*one/352],[1917*one/352,-217*one/352]])
    A=np.eye(2,dtype=object)
    bhat=np.array([180991*one/96132,-17777*one/32044])
    b=np.array([-44709*one/32044,48803*one/96132])
    TSRK['order4']=TwoStepRungeKuttaMethod(d,theta,A,b,Ahat,bhat,type='General')
    #================================================
    d=np.array([-0.210299,-0.0995138])
    theta=-0.186912
    Ahat=np.array([[1.97944,0.0387917],[2.5617,2.3738]])
    A=np.zeros([2,2])
    bhat=np.array([1.45338,0.248242])
    b=np.array([-0.812426,-0.0761097])
    TSRK['order5']=TwoStepRungeKuttaMethod(d,theta,A,b,Ahat,bhat,type='General')
    if which=='All': return TSRK
    else: return TSRK[which]


def load_type2_TSRK(s,p,type='Type II'):
    r"""
    Load a TSRK method from its coefficients in an ASCII file
    (usually from MATLAB).  The coefficients are stored in the
    following order: [s d theta A b Ahat bhat].
    """
    path='/Users/ketch/Research/Projects/MRK/methods/TSRK/'
    file=path+'explicitzr'+str(s)+str(p)+'.tsrk'
    f=open(file,'r')
    coeff=[]
    for line in f:
        for word in line.split():
            coeff.append(float(word))

    s=int(coeff[0])
    A=np.zeros([s,s])
    Ahat=np.zeros([s,s])
    b=np.zeros([s,1])
    bhat=np.zeros([s,1])

    d=np.array(coeff[1:s+1],ndmin=2).T
    theta = coeff[s+1]
    for row in range(s):
        A[row,:] = coeff[s+2+s*row:s+2+s*(row+1)]
    b = np.array(coeff[s**2+s+2:s**2+2*s+2],ndmin=2).T
    for row in range(s):
        Ahat[row,:] = coeff[s**2+2*s+2+s*row:s**2+2*s+2+s*(row+1)]
    bhat = np.array(coeff[2*s**2+2*s+2:2*s**2+3*s+2],ndmin=2).T
    return TwoStepRungeKuttaMethod(d,theta,A,b,Ahat,bhat,type=type)

if __name__ == "__main__":
    import doctest
    doctest.testmod()



"""
Unit tests for NodePy.
This needs to be updated.
"""
from __future__ import absolute_import
import nodepy.linear_multistep_method as lmm
import nodepy.runge_kutta_method as rk
import unittest as ut
from six.moves import range

class LinearMultistepTestCase(ut.TestCase):
    pass

class LMMOrderTest(LinearMultistepTestCase):
    def runTest(self):
        for k in range(1,7):
            ab = lmm.Adams_Bashforth(k)
            self.assertEqual(ab.order(),k)
            am = lmm.Adams_Moulton(k)
            self.assertEqual(am.order(),k+1)
            bdf = lmm.backward_difference_formula(k)
            self.assertEqual(bdf.order(),k)

class LMMSSPCoeffTest(LinearMultistepTestCase):
    def runTest(self):
        for k in range(2,100):
            ssp2 = lmm.elm_ssp2(k)
            self.assertAlmostEqual(ssp2.ssp_coefficient(),(k-2.)/(k-1.),10)

class RungeKuttaTestCase(ut.TestCase):
    def setUp(self):
        self.RKs=rk.loadRKM()

class RKOrderTest(RungeKuttaTestCase):
    knownValues = ( ('FE',1),
            ('SSP22',2),
            ('SSP33',3),
            ('Mid22',2),
            ('RK44',4),
            ('SSP104',4),
            ('GL2',4),
            ('GL3',6),
            ('BuRK65',5) )

    def runTest(self):
        for method, order in self.knownValues:
            self.assertEqual(self.RKs[method].order(),order)

class RKStageOrderTest(RungeKuttaTestCase):
    knownValues = ( ('FE',1),
            ('SSP22',1),
            ('SSP33',1),
            ('Mid22',1),
            ('RK44',1),
            ('SSP104',1),
            ('GL2',2),
            ('GL3',3),
            ('BuRK65',1) )

    def runTest(self):
        for method, stageorder in self.knownValues:
            self.assertEqual(self.RKs[method].stage_order(),stageorder)

class RKAmradTest(RungeKuttaTestCase):
    knownValues = ( ('FE',1),
            ('SSP22',1),
            ('SSP33',1),
            ('Mid22',0),
            ('RK44',0) ,
            ('SSP104',6),
            ('BuRK65',0) )

    def runTest(self):
        for method, SSPCoefficient in self.knownValues:
            self.assertAlmostEqual(self.RKs[method].absolute_monotonicity_radius(),SSPCoefficient,9)

class linAmradTest(RungeKuttaTestCase):
    knownValues = ( ('FE',1),
            ('SSP22',1),
            ('SSP33',1),
            ('Mid22',1),
            ('RK44',1) ,
            ('SSP104',6),
            ('BuRK65',16/9.) )

    def runTest(self):
        for method, R in self.knownValues:
            self.assertAlmostEqual(self.RKs[method].linear_absolute_monotonicity_radius(),R,2)

class RKstabfuntest(RungeKuttaTestCase):
    import sympy
    one = sympy.Rational(1)
    knownValues = ( ('RK44',[one/24,one/6,one/2,one,one]), )
    formula_representation = (('det',True),('lts',False),('pow',False))

    def runTest(self):
        for method, polycoeffs in self.knownValues:
            for formula, use_butcher in self.formula_representation:
                p,q = self.RKs[method].stability_function(formula=formula,use_butcher=use_butcher)
                assert(all(p.coeffs==polycoeffs))
        ssp2 = rk.SSPRK2(3)
        one = self.one
        for formula in ['lts','pow']:
                p,q = ssp2.stability_function(formula=formula,use_butcher=False)
                assert(all(p.coeffs==[one/12,one/2,one,one]))



if __name__== "__main__":
    ut.main()



from __future__ import print_function
from __future__ import division

from __future__ import absolute_import
import numbers
from six.moves import range

#=====================================================
class ODESolver(object):
#=====================================================
    """ Top-level class for numerical ODE solvers """
    def __init__(self):
        pass

    def __step__(self):
        raise NotImplementedError

    def __call__(self,ivp,t0=0,N=5000,dt=None,errtol=None,controllertype='P',
                 x=None,diagnostics=False,use_butcher=False,max_steps=7500):
        """
            Calling an ODESolver numerically integrates the ODE
            u'(t) = f(t,u(t)) with initial value u(0)=u0 from time
            t0 up to time T using the solver.

            The timestep can be controlled in any of three ways:

                1. By specifying N, the total number of steps to take.
                   Then dt = (T-t0)/N.
                2. By specifying dt directly.
                3. For methods with an error estimate (e.g., RK pairs),
                   by specifying an error tolerance.  Then the step
                   size is adjusted using a PI-controller to achieve
                   the requested tolerance.  In this case, dt should
                   also be specified and determines the value of the
                   initial timestep.

            The argument x is used to pass any additional arguments required
            for the RHS function f.

            **Input**:
                - ivp -- An IVP instance (the initial value problem to be solved)
                - t0 -- The initial time from which to integrate
                - N -- The # of steps to take (using a fixed step size)
                - dt -- The step size to use
                - errtol -- The local error tolerance to be observed (using
                  adaptive stepping).  This requires that the method have an
                  error estimator, such as an embedded Runge-Kutta method.
                - controllerType -- The type of adaptive step size control
                  to be used; available options are 'P' and 'PI'.  See
                  :cite:`hairer1993` for details.
                - diagnostics -- if True, return the number of rejected steps
                  and a list of step sizes used, in addition to the solution
                  values and times.

            **Output**:
                - t -- A list of solution times
                - u -- A list of solution values

            If ivp.T is a scalar, the solution is integrated to that time
            and all output step values are returned.  If ivp.T is a list/array,
            the solution is integrated to T[-1] and the solution is returned
            for the times specified in T.

            TODO:

                * Implement an option to not keep all output (for efficiency).
                * Option to keep error estimate history
        """
        # If a list of output times is specified and the method supports dense output,
        # then we don't have to stop exactly at the output times.
        if hasattr(self, 'b_dense') and not isinstance(ivp.T, numbers.Number):
            dense_output = True
        else:
            dense_output = False

        numself = self.__num__()
        f=ivp.rhs; u0=ivp.u0; T=ivp.T
        u=[u0]; t=[t0];
        dthist=[]; errest_hist=[]
        uu = u0
        rejected_steps=0

        if isinstance(T, numbers.Number):
            t_final = T
            t_out = None
        else:
            t_final = T[-1]
            t_out = T

        if t_out is not None:
            iout = 0
            next_output_time = T[iout]
        else:
            next_output_time = t_final

        t_current = t0

        out_now = False

        if not hasattr(self, 'b_dense'):
            if errtol is None:      # Fixed-timestep mode
                if dt is None: dt = (t_final-t0)/float(N)
                dt_standard = dt + 0
                max_steps = max(max_steps, int(round(2*(t_final-t0)/dt)))
                for istep in range(max_steps):

                    if t_current+dt >= next_output_time:
                        dt = next_output_time - t_current
                        out_now = True

                    if not hasattr(self, 'b_dense'):
                        uu = numself.__step__(f,t_current,uu,dt,x=x,use_butcher=use_butcher)
                    else:
                        uu, _ = numself.__step__(f,t_current,uu,dt,[],x=x,use_butcher=use_butcher)

                    t_current += dt
                    if (out_now) or (t_out is None):
                        u.append(uu)
                        t.append(t_current)
                    if t_current >= t_final: break
                    if out_now:
                        iout += 1
                        next_output_time = T[iout]
                        dt = dt_standard

                    out_now = False

            else:                   # Error-control mode
                p=self.embedded_method.p
                alpha = 0.7/p; beta = 0.4/p; kappa = 0.9
                facmin = 0.2; facmax = 5.0
                errestold = errtol
                errest=1.


                for istep in range(max_steps):
                    # Hit next output time exactly:
                    if t_current+dt >= next_output_time: 
                        dt = next_output_time - t_current
                        out_now = True

                    unew,errest = numself.__step__(f,t_current,uu,dt,estimate_error=True,x=x,use_butcher=use_butcher)

                    if errest<=errtol:      # Step accepted
                        t_current += dt
                        if (out_now) or (t_out is None):
                            u.append(unew)
                            t.append(t_current)
                            uu = unew.copy()
                        # Stop if final time reached:
                        if t_current >= t_final: break
                        if out_now:
                            iout += 1
                            next_output_time = T[iout]
                        errestold = errest #Should this happen if step is rejected?
                        dthist.append(dt)
                        errest_hist.append(errest)
                    else: 
                        rejected_steps+=1

                    out_now = False

                    if controllertype=='P':
                      #Compute new dt using P-controller
                      facopt = (errtol/(errest+1.e-6*errtol))**alpha 

                    elif controllertype=='PI':
                      #Compute new dt using PI-controller
                      facopt = ((errtol/errest)**alpha
                                *(errestold/errtol)**beta)
                    else: print('Unrecognized time step controller type')

                    # Set new step size
                    dt = dt * min(facmax,max(facmin,kappa*facopt))

                if istep==max_steps-1:
                    print('Maximum number of steps reached; giving up.')

        else:  # dense output
            if errtol is None:      # Fixed-timestep mode
                if dt is None: dt = (t_final-t0)/float(N)
                dt_standard = dt + 0
                for istep in range(max_steps):

                    thetas = []
                    if t_current+dt >= next_output_time:
                        out_now = True
                        while next_output_time <= t_current + dt:
                            thetas.append( (next_output_time - t_current)/dt )
                            iout += 1
                            if iout >= len(T): break
                            next_output_time = T[iout]

                    uu, output = numself.__step__(f,t_current,uu,dt,thetas,
                                                  x=x,use_butcher=use_butcher)

                    if output:
                        for i, outsol in enumerate(output):
                            u.append(outsol)
                            t.append(t_current + dt*thetas[i])
                    t_current += dt
                    if t_current >= t_final: break

                    out_now = False

            else:                   # Error-control mode
                raise NotImplementedError

        if diagnostics==False:
            return t, u
        else:
            if errtol is None:
                return t,u,rejected_steps,dthist
            else:
                return t,u,rejected_steps,dthist,errest_hist



"""
Routines for plotting Runge-Kutta method stage dependency graphs.

The stage dependency graph shows the order in which stages must be computed.
For traditional methods, this is not very interesting, but for special classes
of methods this reveals the potential for parallelism.

**Examples**::

    >>> from nodepy import rk
    >>> ex4 = rk.extrap(4)
    >>> plot_dependency_graph(ex4)
"""
from __future__ import absolute_import
def plot_dependency_graph(rkm,remove_edges=False):
    """Plot the stage dependency graph of a Runge-Kutta method.
    
    **Inputs**:

        - rkm : A RungeKuttaMethod
    """
    import networkx as nx
    import numpy as np
    G = rk_dependency_graph(rkm)
    lp = _longest_paths(G)

    hpos = [0]*(max(lp)+1)
    pos = {}
    for node in G.nodes():
        hpos[lp[node]]+=1
        pos[node] = np.array([hpos[lp[node]],lp[node]])

    if remove_edges:
       _remove_extra_edges(G)

    nx.draw_networkx(G,pos)

def rk_dependency_graph(rkm):
    """Construct the dependency graph of the stages of a Runge-Kutta method."""
    import networkx as nx
    import numpy as np
    s = len(rkm)
    K = np.zeros((s+1,s+1))
    K[:,:-1] = np.vstack((rkm.A,rkm.b))
    G = nx.from_numpy_matrix(np.abs(K.T)>0,nx.DiGraph())
    return G

def _longest_paths(G):
    """Return a dict of nodes with values corresponding to longest paths from node"""
    lp = {}
    lp_old = lp.copy()
    for node in G.nodes():
        lp[node] = 0

    while lp != lp_old:
        lp_old = lp.copy()
        for edge in G.edges():
            lp[edge[1]] = max(lp[edge[1]],lp[edge[0]]+1)

    return lp

def _remove_extra_edges(G):
    """Remove edges that aren't needed in establishing ordering."""
    lp = _longest_paths(G)
    to_remove = []
    for edge in G.edges():
        if lp[edge[0]]<(lp[edge[1]]-1):
            to_remove.append(edge)
    for edge in to_remove:
        G.remove_edge(edge[0],edge[1])


if __name__ == "__main__":
    import doctest
    doctest.testmod()



"""
**Examples**::

    >>> import nodepy.linear_multistep_method as lm
    >>> ab3=lm.Adams_Bashforth(3)
    >>> ab3.order()
    3
    >>> bdf2=lm.backward_difference_formula(2)
    >>> bdf2.order()
    2
    >>> bdf2.is_zero_stable()
    True
    >>> bdf7=lm.backward_difference_formula(7)
    >>> bdf7.is_zero_stable()
    False
    >>> bdf3=lm.backward_difference_formula(3)
    >>> bdf3.A_alpha_stability()
    86
    >>> ssp32=lm.elm_ssp2(3)
    >>> ssp32.order()
    2
    >>> ssp32.ssp_coefficient()
    1/2
    >>> ssp32.plot_stability_region() #doctest: +ELLIPSIS
    <Figure...

"""
from __future__ import print_function

from __future__ import absolute_import
import numpy as np
import copy
import sympy
import nodepy.snp as snp
import matplotlib.pyplot as plt
from sympy import symbols, latex
from nodepy.snp import printable
from nodepy.general_linear_method import GeneralLinearMethod
from six.moves import map
from six.moves import range
from sympy import Rational
try:
    import sympy.combinatorial as combinatorial
except ImportError:
    import sympy.functions.combinatorial as combinatorial


class LinearMultistepMethod(GeneralLinearMethod):
    r"""
        Implementation of linear multistep methods in the form:

        `\alpha_k y_{n+k} + \alpha_{k-1} y_{n+k-1} + ... + \alpha_0 y_n
        = h ( \beta_k f_{n+k} + ... + \beta_0 f_n )`

        Methods are automatically normalized so that \alpha_k=1.

        Notes: Representation follows Hairer & Wanner p. 368, NOT Butcher.

        **References**:
            * :cite:`hairer1993` Chapter III
            * :cite:`butcher2003`
    """
    def __init__(self,alpha,beta,name='Linear multistep method',shortname='LMM',
                 description=''):
        self.beta  = beta /alpha[-1]
        self.alpha = alpha/alpha[-1]
        self.name = name
        self.shortname = shortname
        if description is not '':
            self.info = description
        else:
            if self.is_explicit():
                exp_str = "Explicit"
            else:
                exp_str = "Implicit"

            self.info = "%s %d-step method of order %d" % (exp_str, len(self), self.order())

    def __num__(self):
        """
        Returns a copy of the method but with floating-point coefficients.
        This is useful whenever we need to operate numerically without
        worrying about the representation of the method.
        """
        numself = copy.deepcopy(self)
        if self.alpha.dtype==object:
            numself.alpha=np.array(self.alpha,dtype=np.float64)
            numself.beta=np.array(self.beta,dtype=np.float64)
        return numself

    def characteristic_polynomials(self):
        r"""
        Returns the characteristic polynomials (also known as generating
        polynomials) of a linear multistep method.  They are:

        `\rho(z) = \sum_{j=0}^k \alpha_k z^k`

        `\sigma(z) = \sum_{j=0}^k \beta_k z^k`

        **Examples**::

            >>> from nodepy import lm
            >>> ab5 = lm.Adams_Bashforth(5)
            >>> rho,sigma = ab5.characteristic_polynomials()
            >>> print(rho)
               5     4
            1 x - 1 x

            >>> print(sigma)
                  4         3         2
            2.64 x - 3.853 x + 3.633 x - 1.769 x + 0.3486

        **Reference**: :cite:`hairer1993` p. 370, eq. 2.4

        """
        rho=np.poly1d(self.alpha[::-1])
        sigma=np.poly1d(self.beta[::-1])
        return rho, sigma

    def order(self,tol=1.e-10):
        r""" Return the order of the local truncation error of a linear multistep method.

        **Examples**::

            >>> from nodepy import lm
            >>> am3=lm.Adams_Moulton(3)
            >>> am3.order()
            4
        """
        p = 0
        while True:
            if self._satisfies_order_conditions(p+1,tol):
                p = p + 1
            else:
                return p

    def _satisfies_order_conditions(self,p,tol):
        """ Return True if the linear multistep method satisfies
            the conditions of order p (only) """
        ii=snp.arange(len(self.alpha))
        return abs(sum(ii**p*self.alpha-p*self.beta*ii**(p-1)))<tol

    def absolute_monotonicity_radius(self):
        return self.ssp_coefficient()

    @property
    def p(self):
        return self.order()

    def latex(self):
        r"""
        Print a LaTeX representation of a linear multistep formula.

        **Example**::

            >>> from nodepy import lm
            >>> print(lm.Adams_Bashforth(2).latex())
            \begin{align} y_{n + 2} - y_{n + 1} = \frac{3}{2}h f(y_{n + 1}) - \frac{1}{2}h f(y_{n})\end{align}

        """
        n = symbols('n')
        k = len(self)
        alpha_terms = []
        beta_terms = []
        for i in range(k+1):
            subscript = latex(n+k-i)
            if self.alpha[k-i] != 0:
                alpha_terms.append(printable(self.alpha[k-i],return_one=False) +
                                   ' y_{'+subscript+'}')
            if self.beta[k-i] != 0:
                beta_terms.append(printable(self.beta[k-i],return_one=False) +
                                  'h f(y_{'+subscript+'})')
        lhs = ' + '.join(alpha_terms)
        rhs = ' + '.join(beta_terms)
        s = r'\begin{align}'+ ' = '.join([lhs,rhs]) + r'\end{align}'
        s = s.replace('+ -','-')
        return s

    def ssp_coefficient(self):
        r""" Return the SSP coefficient of the method.

         The SSP coefficient is given by

        `\min_{0 \le j < k} -\alpha_k/beta_k`

        if `\alpha_j<0` and `\beta_j>0` for all `j`, and is equal to
        zero otherwise.


        **Examples**::

            >>> from nodepy import lm
            >>> ssp32=lm.elm_ssp2(3)
            >>> ssp32.ssp_coefficient()
            1/2

            >>> bdf2=lm.backward_difference_formula(2)
            >>> bdf2.ssp_coefficient()
            0
        """
        if np.any(self.alpha[:-1]>0) or np.any(self.beta<0):
            return 0

        return min([-self.alpha[j]/self.beta[j]
                    for j in range(len(self.alpha)-1) if self.beta[j]!=0])

    def plot_stability_region(self,N=100,bounds=None,color='r',filled=True, alpha=1.,
                              to_file=False,longtitle=False):
        r"""
            The region of absolute stability of a linear multistep method is
            the set

            `\{ z \in C : \rho(\zeta) - z \sigma(zeta) \text{ satisfies the root condition} \}`

            where `\rho(zeta)` and `\sigma(zeta)` are the characteristic
            functions of the method.

            Also plots the boundary locus, which is
            given by the set of points z:

            `\{z | z=\rho(\exp(i\theta))/\sigma(\exp(i\theta)), 0\le \theta \le 2*\pi \}`

            Here `\rho` and `\sigma` are the characteristic polynomials
            of the method.

            Reference: :cite:`leveque2007` section 7.6.1


            **Input**: (all optional)
                - N       -- Number of gridpoints to use in each direction
                - bounds  -- limits of plotting region
                - color   -- color to use for this plot
                - filled  -- if true, stability region is filled in (solid); otherwise it is outlined
        """
        rho, sigma = self.__num__().characteristic_polynomials()
        mag = lambda z: _root_condition(rho-z*sigma)
        vmag = np.vectorize(mag)

        z = self._boundary_locus()
        if bounds is None:
            # Use boundary locus to decide plot region
            realmax, realmin = np.max(np.real(z)), np.min(np.real(z))
            imagmax, imagmin = np.max(np.imag(z)), np.min(np.imag(z))
            deltar, deltai = realmax-realmin, imagmax-imagmin
            bounds = (realmin-deltar/5.,realmax+deltar/5.,
                      imagmin-deltai/5.,imagmax+deltai/5.)

        y=np.linspace(bounds[2],bounds[3],N)
        Y=np.tile(y[:,np.newaxis],(1,N))
        x=np.linspace(bounds[0],bounds[1],N)
        X=np.tile(x,(N,1))
        Z=X+Y*1j

        R=1.5-vmag(Z)

        if filled:
            plt.contourf(X,Y,R,[0,1],colors=color,alpha=alpha)
        else:
            plt.contour(X,Y,R,[0,1],colors=color,alpha=alpha)

        fig = plt.gcf()
        ax = fig.get_axes()
        if longtitle:
            plt.setp(ax,title='Absolute Stability Region for '+self.name)
        else:
            plt.setp(ax,title='Stability region')

        plt.plot([0,0],[bounds[2],bounds[3]],'--k',linewidth=2)
        plt.plot([bounds[0],bounds[1]],[0,0],'--k',linewidth=2)
        plt.plot(np.real(z),np.imag(z),color='k',linewidth=3)
        plt.axis(bounds)
        plt.axis('image')

        if to_file:
            plt.savefig(to_file, transparent=True, bbox_inches='tight', pad_inches=0.3)

        plt.draw()
        return fig

    def plot_boundary_locus(self,N=1000,figsize=None):
        r"""Plot the boundary locus, which is
            given by the set of points

            `\{z | z=\rho(\exp(i\theta))/\sigma(\exp(i\theta)), 0\le \theta \le 2*\pi \}`

            where `\rho` and `\sigma` are the characteristic polynomials
            of the method.

            Reference: :cite:`leveque2007` section 7.6.1
        """
        z = self._boundary_locus(N)

        if figsize is None:
            plt.figure()
        else:
            plt.figure(figsize=figsize)
        plt.plot(np.real(z),np.imag(z),color='k',linewidth=3)
        plt.axis('image')
        bounds = plt.axis()
        plt.plot([0,0],[bounds[2],bounds[3]],'--k',linewidth=2)
        plt.plot([bounds[0],bounds[1]],[0,0],'--k',linewidth=2)
        plt.title('Boundary locus for '+self.name)
        plt.draw()

    def _boundary_locus(self, N=1000):
        r"""Compute the boundary locus, which is
            given by the set of points

            `\{z | z=\rho(\exp(i\theta))/\sigma(\exp(i\theta)), 0\le \theta \le 2*\pi \}`

            where `\rho` and `\sigma` are the characteristic polynomials
            of the method.

            Reference: :cite:`leveque2007` section 7.6.1
        """
        theta=np.linspace(0.,2*np.pi,N)
        zeta = np.exp(theta*1j)
        rho,sigma=self.__num__().characteristic_polynomials()
        z = rho(zeta)/sigma(zeta)

        return z

    def A_alpha_stability(self, N=1000, tol=1.e-14):
        r"""Angle of `A(\alpha)`-stability.

        The result is given in degrees.  The result is only accurate to
        about 1 degree, so we round down.

        **Examples**:

            >>> from nodepy import lm
            >>> bdf5 = lm.backward_difference_formula(5)
            >>> bdf5.A_alpha_stability()
            51
        """
        from math import atan2, floor

        z = self._boundary_locus(N)
        rad = list(map(atan2,np.imag(z),np.real(z)))
        rad = np.mod(np.array(rad),2*np.pi)

        return min(int(floor(np.min(np.abs(np.where(np.real(z)<-tol,rad,1.e99)-np.pi))/np.pi*180)),90)

    def is_explicit(self):
        return self.beta[-1]==0

    def is_zero_stable(self,tol=1.e-13):
        r""" True if the method is zero-stable.

        **Examples**::

            >>> from nodepy import lm
            >>> bdf5=lm.backward_difference_formula(5)
            >>> bdf5.is_zero_stable()
            True
        """
        rho, sigma = self.characteristic_polynomials()
        return _root_condition(rho,tol)

    def __len__(self):
        r"""Returns the number of steps used."""
        return len(self.alpha)-1


class AdditiveLinearMultistepMethod(GeneralLinearMethod):
    r"""
        Method for solving equations of the form

        `y'(t) = f(y) + g(y)`

        The method takes the form

        `\alpha_k y_{n+k} + \alpha_{k-1} y_{n+k-1} + ... + \alpha_0 y_n
        = h ( \beta_k f_{n+k} + ... + \beta_0 f_n
        + \gamma_k f_{n+k} + ... + \gamma_0 f_n )`

        Methods are automatically normalized so that \alpha_k=1.

        The usual reference for these is Ascher, Ruuth, and Whetton.
        But we follow a different notation (as just described).
    """
    def __init__(self, alpha, beta, gamma, name='Additive linear multistep method'):
        self.beta  = beta /alpha[-1]
        self.gamma = gamma/alpha[-1]
        self.alpha = alpha/alpha[-1]
        self.name = name
        self.method1 = LinearMultistepMethod(alpha, beta)
        self.method2 = LinearMultistepMethod(alpha, gamma)

    def __num__(self):
        """
        Returns a copy of the method but with floating-point coefficients.
        This is useful whenever we need to operate numerically without
        worrying about the representation of the method.
        """
        numself = copy.deepcopy(self)
        if self.alpha.dtype == object:
            numself.alpha = np.array(self.alpha, dtype=np.float64)
            numself.beta  = np.array(self.beta, dtype=np.float64)
            numself.gamma = np.array(self.gamma, dtype=np.float64)
        return numself

    def order(self,tol=1.e-10):
        r""" Return the order of the local truncation error of an additive
             linear multistep method.  The output is the minimum of the
             order of the component methods.
        """
        orders = []
        for method in (self.method1,self.method2):
            p = 0
            while True:
                if method._satisfies_order_conditions(p+1,tol):
                    p = p + 1
                else:
                    orders.append(p)
                    break

        return min(orders)

    def plot_imex_stability_region(self,both_real=False,N=100,color='r',filled=True,
                                   alpha=1.,fignum=None,bounds=[-10, 1, -5, 5]):
        r"""
            **Input**: (all optional)
                - N       -- Number of gridpoints to use in each direction
                - bounds  -- limits of plotting region
                - color   -- color to use for this plot
                - filled  -- if true, stability region is filled in (solid); otherwise it is outlined
        """
        rho, sigma1 = self.method1.__num__().characteristic_polynomials()
        rho, sigma2 = self.method2.__num__().characteristic_polynomials()
        if both_real:
            mag = lambda a, b: _max_root(rho - a*sigma1 - b*sigma2)
        else:
            mag = lambda a, b: _max_root(rho - a*sigma1 - 1j*b*sigma2)
        vmag = np.vectorize(mag)

        y = np.linspace(bounds[2],bounds[3],N)
        Y = np.tile(y[:,np.newaxis],(1,N))
        x = np.linspace(bounds[0],bounds[1],N)
        X = np.tile(x,(N,1))

        R = vmag(X,Y)

        h = plt.figure(fignum)
        if filled:
            plt.contourf(X,Y,R,[0,1],colors=color,alpha=alpha)
        else:
            plt.contour(X,Y,R,[0,1],colors=color,alpha=alpha)
        plt.contour(X,Y,R,np.linspace(0,1,10),colors='k')
        plt.title('IMEX Stability Region for '+self.name)
        plt.plot([0,0],[bounds[2],bounds[3]],'--k',linewidth=2)
        plt.plot([bounds[0],bounds[1]],[0,0],'--k',linewidth=2)
        plt.axis(bounds)
        return h

    def stiff_damping_factor(self,epsilon=1.e-7):
        r"""
        Return the magnitude of the largest root at z=-inf.
        This routine just computes a numerical approximation
        to the true value (with absolute accuracy epsilon).
        """
        rho, sigma1 = self.method1.__num__().characteristic_polynomials()
        rho, sigma2 = self.method2.__num__().characteristic_polynomials()
        mag = lambda a, b: _max_root(rho - a*sigma1 - 1j*b*sigma2)

        f=[]
        z=-1.
        f.append(mag(z,0))
        while True:
            z = z*10.
            f.append(mag(z,0))
            if np.abs(f[-1]-f[-2]) < epsilon:
                return f[-1]
            if len(f)>100:
                print(f)
                raise Exception('Unable to compute stiff damping factor: slow convergence')


# ======================================================
def _max_root(p):
    return max(np.abs(p.r))

def _root_condition(p,tol=1.e-13):
    r""" True if the polynomial `p` has all roots inside
    the unit circle and roots on the boundary of the unit circle
    are simple.

    **Examples**::

        >>> from nodepy import lm
        >>> p = np.poly1d((1,0.4,2,0.5))
        >>> lm._root_condition(p)
        False
    """
    if max(np.abs(p.r))>(1+tol):
        return False

    mod_one_roots = [r for r in p.r if abs(abs(r)-1)<tol]
    for i,r1 in enumerate(mod_one_roots):
        for r2 in mod_one_roots[i+1:]:
            if abs(r1-r2)<tol:
                return False
    return True


# ======================================================
# Families of multistep methods
# ======================================================

def Adams_Bashforth(k):
    r"""
    Construct the k-step, Adams-Bashforth method.
    The methods are explicit and have order k.
    They have the form:

    `y_{n+1} = y_n + h \sum_{j=0}^{k-1} \beta_j f(y_n-k+j+1)`

    They are generated using equations (1.5) and (1.7) from
    :cite:`hairer1993` III.1, along with the binomial expansion.

    **Examples**::

        >>> import nodepy.linear_multistep_method as lm
        >>> ab3=lm.Adams_Bashforth(3)
        >>> ab3.order()
        3

        Reference: :cite:`hairer1993`
    """
    one = Rational(1,1)

    alpha=snp.zeros(k+1)
    beta=snp.zeros(k+1)
    alpha[k]=one
    alpha[k-1]=-one
    gamma=snp.zeros(k)
    gamma[0]=one
    beta[k-1]=one
    betaj=snp.zeros(k+1)
    for j in range(1,k):
        gamma[j]=one-sum(gamma[:j]/snp.arange(j+1,1,-1))
        for i in range(0,j+1):
            betaj[k-i-1]=(-one)**i*combinatorial.factorials.binomial(j,i)*gamma[j]
        beta=beta+betaj
    name=str(k)+'-step Adams-Bashforth'
    return LinearMultistepMethod(alpha,beta,name=name,shortname='AB'+str(k))

def Nystrom(k):
    r"""
    Construct the k-step explicit Nystrom linear multistep method.
    The methods are explicit and have order k.
    They have the form:

    `y_{n+1} = y_{n-1} + h \sum_{j=0}^{k-1} \beta_j f(y_n-k+j+1)`

    They are generated using equations (1.13) and (1.7) from
    :cite:`hairer1993` III.1, along with the binomial expansion
    and the relation in exercise 4 on p. 367.

    Note that the term "Nystrom method" is also commonly used to refer
    to a class of methods for second-order ODEs; those are NOT
    the methods generated by this function.

    **Examples**::

        >>> import nodepy.linear_multistep_method as lm
        >>> nys3=lm.Nystrom(6)
        >>> nys3.order()
        6

        Reference: :cite:`hairer1993`
    """
    one = Rational(1,1)

    alpha = snp.zeros(k+1)
    alpha[k] = one
    alpha[k-2] = -one

    beta  = snp.zeros(k+1)
    kappa = snp.zeros(k)
    gamma = snp.zeros(k)
    gamma[0]  =   one
    kappa[0]  = 2*one
    beta[k-1] = 2*one
    betaj = snp.zeros(k+1)
    for j in range(1,k):
        gamma[j] = one-sum(gamma[:j]/snp.arange(j+1,1,-1))
        kappa[j] = 2 * gamma[j] - gamma[j-1]
        for i in range(0,j+1):
            betaj[k-i-1] = (-one)**i*combinatorial.factorials.binomial(j,i)*kappa[j]
        beta = beta+betaj
    name = str(k)+'-step Nystrom'
    return LinearMultistepMethod(alpha,beta,name=name,shortname='Nys'+str(k))

def Adams_Moulton(k):
    r"""
        Construct the k-step, Adams-Moulton method.
        The methods are implicit and have order k+1.
        They have the form:

        `y_{n+1} = y_n + h \sum_{j=0}^{k} \beta_j f(y_n-k+j+1)`

        They are generated using equation (1.9) and the equation in
        Exercise 3 from Hairer & Wanner III.1, along with the binomial
        expansion.

        **Examples**::

            >>> import nodepy.linear_multistep_method as lm
            >>> am3=lm.Adams_Moulton(3)
            >>> am3.order()
            4

        Reference: :cite:`hairer1993`
    """

    alpha=snp.zeros(k+1)
    beta=snp.zeros(k+1)
    alpha[k]=1
    alpha[k-1]=-1
    gamma=snp.zeros(k+1)
    gamma[0]=1
    beta[k]=1
    betaj=snp.zeros(k+1)
    for j in range(1,k+1):
        gamma[j]= -sum(gamma[:j]/snp.arange(j+1,1,-1))
        for i in range(0,j+1):
            betaj[k-i]=(-1)**i*combinatorial.factorials.binomial(j,i)*gamma[j]
        beta=beta+betaj
    name=str(k)+'-step Adams-Moulton'
    return LinearMultistepMethod(alpha,beta,name=name,shortname='AM'+str(k))

def Milne_Simpson(k):
    r"""
        Construct the k-step, Milne-Simpson method.
        The methods are implicit and (for k>=3) have order k+1.
        They have the form:

        `y_{n+1} = y_{n-1} + h \sum_{j=0}^{k} \beta_j f(y_n-k+j+1)`

        They are generated using equation (1.15), the equation in
        Exercise 3, and the relation in exercise 4, all from Hairer & Wanner
        III.1, along with the binomial expansion.

        **Examples**::

            >>> import nodepy.linear_multistep_method as lm
            >>> ms3=lm.Milne_Simpson(3)
            >>> ms3.order()
            4

        Reference: :cite:`hairer1993`
    """

    alpha = snp.zeros(k+1)
    beta  = snp.zeros(k+1)
    alpha[k] = 1
    alpha[k-2] = -1
    gamma = snp.zeros(k+1)
    kappa = snp.zeros(k+1)
    gamma[0] = 1
    kappa[0] = 2
    beta[k]  = 2
    betaj = snp.zeros(k+1)
    for j in range(1,k+1):
        gamma[j] = -sum(gamma[:j]/snp.arange(j+1,1,-1))
        kappa[j] = 2 * gamma[j] - gamma[j-1]
        for i in range(0,j+1):
            betaj[k-i] = (-1)**i*combinatorial.factorials.binomial(j,i)*kappa[j]
        beta = beta+betaj
    name = str(k)+'-step Milne-Simpson'
    return LinearMultistepMethod(alpha,beta,name=name,shortname='MS'+str(k))

def backward_difference_formula(k):
    r"""
        Construct the k-step backward differentiation method.
        The methods are implicit and have order k.
        They have the form:

        `\sum_{j=0}^{k} \alpha_j y_{n+k-j+1} = h \beta_j f(y_{n+1})`

        They are generated using equation (1.22') from Hairer & Wanner III.1,
            along with the binomial expansion.

        **Examples**::

            >>> import nodepy.linear_multistep_method as lm
            >>> bdf4=lm.backward_difference_formula(4)
            >>> bdf4.A_alpha_stability()
            73

        **Reference**: :cite:`hairer1993` pp. 364-365
    """

    alpha=snp.zeros(k+1)
    beta=snp.zeros(k+1)
    beta[k]=1
    gamma=snp.zeros(k+1)
    gamma[0]=1
    alphaj=snp.zeros(k+1)
    for j in range(1,k+1):
        gamma[j]= sympy.Rational(1,j)
        for i in range(0,j+1):
            alphaj[k-i]=(-1)**i*combinatorial.factorials.binomial(j,i)*gamma[j]
        alpha=alpha+alphaj
    name=str(k)+'-step BDF'
    return LinearMultistepMethod(alpha,beta,name=name,shortname='BDF'+str(k))

def elm_ssp2(k):
    r"""
    Returns the optimal SSP k-step linear multistep method of order 2.

    **Examples**::

        >>> import nodepy.linear_multistep_method as lm
        >>> lm10=lm.elm_ssp2(10)
        >>> lm10.ssp_coefficient()
        8/9
    """
    alpha=snp.zeros(k+1)
    beta=snp.zeros(k+1)
    alpha[-1]=sympy.Rational(1,1)
    alpha[0]=sympy.Rational(-1,(k-1)**2)
    alpha[k-1]=sympy.Rational(-(k-1)**2+1,(k-1)**2)
    beta[k-1]=sympy.Rational(k,k-1)
    name='Optimal '+str(k)+'-step, 2nd order SSP method.'
    return LinearMultistepMethod(alpha,beta,name=name)

def sand_cc(s):
    r""" Construct Sand's circle-contractive method of order `p=2(s+1)`
         that uses `2^s + 1` steps.

    **Examples**::

        >>> import nodepy.linear_multistep_method as lm
        >>> cc4 = lm.sand_cc(4)
        >>> cc4.order()
        10
        >>> cc4.ssp_coefficient()
        1/8

    **Reference**: :cite:`sand1986`
    """
    one  = sympy.Rational(1)
    zero = sympy.Rational(0)

    k = 2**s + 1

    Jn = [k,k-1]
    for i in range(1,s+1):
        Jn.append(k-1-2**i)

    alpha = snp.zeros(k+1)
    beta  = snp.zeros(k+1)

    # This is inefficient
    for j in Jn:
        tau_product = one
        tau_sum = zero
        tau = [one/(j-i) for i in Jn if i!=j]
        tau_product = np.prod(tau)
        tau_sum = np.sum(tau)
        beta[j] = tau_product**2
        alpha[j] = 2*beta[j]*tau_sum
    return LinearMultistepMethod(alpha,beta,'Sand circle-contractive')

def arw2(gam,c):
    r"""Returns the second order IMEX additive multistep method based on the
    parametrization in Section 3.2 of Ascher, Ruuth, & Whetton.  The parameters
    are gam and c.  Known methods are obtained with the following values:

        (1/2,0):   CNAB
        (1/2,1/8): MCNAB
        (0,1):     CNLF
        (1,0):     SBDF

    **Examples**::

        >>> from nodepy import lm
        >>> import sympy
        >>> CNLF = lm.arw2(0,sympy.Rational(1))
        >>> CNLF.order()
        2
        >>> CNLF.method1.ssp_coefficient()
        1
        >>> CNLF.method2.ssp_coefficient()
        0
        >>> print(CNLF.stiff_damping_factor()) #doctest: +ELLIPSIS
        0.999...
    """
    half = sympy.Rational(1,2)
    alpha = snp.array([gam-half,-2*gam,gam+half])
    beta  = snp.array([c/2,1-gam-c,gam+c/2])  # implicit part
    gamma = snp.array([-gam,gam+1,0])  # explicit part
    return AdditiveLinearMultistepMethod(alpha,beta,gamma,'ARW2('+str(gam)+','+str(c)+')')

def arw3(gam,theta,c):
    r"""Returns the third order IMEX additive multistep method based on the
    parametrization in Section 3.3 of Ascher, Ruuth, & Whetton.  The parameters
    are gamma, theta, and c.  Known methods are obtained with the following values:

        (1,0,0):     SBDF3

    Note that there is one sign error in the ARW paper; it is corrected here.
    """
    half = sympy.Rational(1,2)
    third = sympy.Rational(1,3)
    alpha = snp.array([-half*gam**2+third/2, 3*half*gam**2+gam-1, -3*half*gam**2-2*gam+half-theta,
                       half*gam**2+gam+third+theta])
    beta  = snp.array([5*half/6*theta-c, (gam**2-gam)*half+3*c-4*theta*third, 1-gam**2-3*c+23*theta*third/4,
                       (gam**2+gam)*half+c])
    gamma = snp.array([(gam**2+gam)*half+5*theta*third/4, -gam**2-2*gam-4*theta*third,
                       (gam**2+3*gam)*half+1+23*theta*third/4,0])
    return AdditiveLinearMultistepMethod(alpha,beta,gamma,'ARW3('+str(gam)+','+str(theta)+','+str(c)+')')


def loadLMM(which='All'):
    """
    Load a set of standard linear multistep methods for testing.

    **Examples**::

        >>> from nodepy import lm
        >>> ebdf5 = lm.loadLMM('eBDF5')
        >>> ebdf5.is_zero_stable()
        True
    """
    LM={}
    # ================================================
    alpha = snp.array([-12,75,-200,300,-300,137])/sympy.Rational(137,1)
    beta = snp.array([60,-300,600,-600,300,0])/sympy.Rational(137,1)
    LM['eBDF5'] = LinearMultistepMethod(alpha,beta,'eBDF 5')
    # ================================================
    theta = sympy.Rational(1,2)
    alpha = snp.array([-1,1])
    beta = snp.array([1-theta,theta])
    gamma = snp.array([1,0])
    LM['ET112'] = AdditiveLinearMultistepMethod(alpha,beta,gamma,'Euler-Theta')
    # ================================================
    if which=='All':
        return LM
    else:
        return LM[which]


if __name__ == "__main__":
    import doctest
    doctest.testmod()



"""
The principal objects in NodePy are ODE solvers. The object upon which a solver
acts is an initial value problem. Mathematically, an initial value problem
(IVP) consists of one or more ordinary differential equations and an initial
condition:

\\begin{align*} u(t) & = F(u) & u(0) & = u_0. \\end{align*}

This module implements the initial value problem as a class.

**Examples**::

    >>> from nodepy import ivp, rk
    >>> myivp = ivp.detest('A1')
    >>> myivp
    Problem Name:  A1
    Description:   Problem A1 of the non-stiff DETEST suite.

    # Integrate this problem with a Runge-Kutta method
    >>> rk4 = rk.loadRKM('RK44')
    >>> t, y = rk4(myivp)
    >>> y[-1][0] # doctest: +ELLIPSIS
    2.06115362252...e-09

    # Load a large set of non-stiff problems
    >>> problems = detest_suite_plus()

    # Load a few stiff problems
    >>> problems = detest_stiff_suite()
"""
from __future__ import print_function

from __future__ import absolute_import
import numpy as np
from six.moves import range

class IVP(object):
    """
        In NodePy, an initial value problem is an object with the following
        properties:

            rhs: The right-hand-side function; i.e. F where u'=F(u).
            u0:  The initial condition.
            T:   The (default) final time of solution.

        Optionally an IVP may possess the following:
            exact: a function that takes one argument (t) and returns
                   the exact solution (Should we make this a function of
                   u0 as well?)
            dt0: The default initial timestep when a variable step size
                 integrator is used.
            Any other problem-specific parameters.

    """
    def __init__(self, f=None, u0=1., t0=0., T=1., exact=None, desc='', name=''):
        self.u0  = u0
        self.rhs = f
        self.T   = T
        self.exact = exact
        self.description = desc
        self.t0 = t0
        self.dt0 = 1.e-2 # This doesn't really belong here
        self.name = name

    def __repr__(self):
        return 'Problem Name:  '+self.name+'\n'+'Description:   '+self.description

def load_ivp(ivpname='All'):
    r"""Load some very simple initial value problems.  The following methods
        are available::

        >>> from nodepy import ivp
        >>> ivps = ivp.load_ivp('all')
        >>> for problem in sorted(ivps.values(), key=lambda x: x.name):
        ...     print("{}: {}".format(problem.name, problem.description).rstrip())
        2odes:
        nlsin: A simple nonlinear scalar problem
        ode1:
        ode2:
        test: Dahlquist's test problem; $f(y) = \lambda y$
        vdp: Van der Pol oscillator with epsilon = 1/10.
        zoltan: The linear scalar test problem with abs value
    """
    ivps = {}

    name = 'test'
    description = r"Dahlquist's test problem; $f(y) = \lambda y$"
    u0 = 1.
    rhs = lambda t,u: u
    exact = lambda t, u0=u0: u0*np.exp(t)
    ivps[name] = IVP(f=rhs, u0=u0, T=5., desc=description, exact=exact, name=name)

    name = 'zoltan'
    u0 = 1.
    rhs = lambda t,u: -100*abs(u)
    exact = lambda t, u0=u0: u0*np.exp(-100*t)
    description = 'The linear scalar test problem with abs value'
    ivps[name] = IVP(f=rhs, u0=u0, T=10., desc=description, exact=exact, name=name)

    name = 'nlsin'
    u0 = 1.
    rhs = lambda t,u: 4.*u*float(np.sin(t))**3*np.cos(t)
    exact = lambda t, u0=u0: u0*np.exp((np.sin(t))**4)
    description = 'A simple nonlinear scalar problem'
    ivps[name] = IVP(f=rhs, u0=u0, T=5., desc=description, exact=exact, name=name)

    name = 'ode1'
    rhs = lambda t,u: 4.*t*np.sqrt(u)
    exact = lambda t: (1.+t**2)**2
    ivps[name] = IVP(f=rhs, u0=1., T=5., exact=exact, name=name)

    name = 'ode2'
    u0=np.exp(1.)
    rhs = lambda t,u: u/t*np.log(u)
    exact = lambda t: np.exp(2.*t)
    ivps[name] = IVP(f=rhs, u0=u0, t0=0.5, T=5., exact=exact, name=name)

    name = '2odes'
    u0=np.array([1.,1.])
    rhs = lambda t,u: np.array([u[0], 2.*u[1]])
    exact = lambda t: np.array([np.exp(t), np.exp(2.*t)])
    ivps[name] = IVP(f=rhs, u0=u0, T=5., exact=exact, name=name)

    name = 'vdp'
    eps=0.1
    u0=np.array([2.,-0.65])
    rhs = lambda t,u: np.array([u[1], 1./eps*(-u[0]+(1.-u[0]**2)*u[1])])
    description = 'Van der Pol oscillator with epsilon = 1/10.'
    ivps[name] = IVP(f=rhs, u0=u0, T=5., exact=exact, name=name, desc=description)

    if ivpname.lower() == 'all':
        return ivps
    else:
        return ivps[ivpname]


def detest(testkey):
    """
        Load problems from the non-stiff DETEST problem set.
        The set consists of six groups of problems, as follows:

            * A1-A5 -- Scalar problems
            * B1-B5 -- Small systems (2-3 equations)
            * C1-C5 -- Moderate size systems (10-50 equations)
            * D1-D5 -- Orbit equations with varying eccentricities
            * E1-E5 -- Second order equations
            * F1-F5 -- Problems with discontinuities
	    * SB1-SB3 -- Periodic Orbit problem from Shampine Baca paper pg.11,13

        .. note::
            Although this set of problems was not intended to become a
            standard, and although there are certain dangers in accepting
            any particular set of problems as a universal standard, it
            is nevertheless sometimes useful to try a new method on this
            test set due to the availability of published results for
            many existing methods.

        Reference: :cite:`enright1987`.
    """
    import numpy as np
    ivp=IVP()
    if testkey=='A1':
        ivp.u0=1.
        ivp.T=20.
        ivp.rhs = lambda t,u: -u
        ivp.dt0 = 1.e-2
    elif testkey=='A2':
        ivp.u0=1.
        ivp.T=20.
        ivp.rhs = lambda t,u: -0.5*u**3
        ivp.dt0 = 1.e-2
    elif testkey=='A3':
        ivp.u0=1.
        ivp.T=20.
        ivp.rhs = lambda t,u: u*np.cos(t)
        ivp.dt0 = 1.e-2
    elif testkey=='A4':
        ivp.u0=1.
        ivp.T=20.
        ivp.rhs = lambda t,u: 0.25*u*(1.-0.05*u)
        ivp.dt0 = 1.e-5
    elif testkey=='A5':
        ivp.u0=4.
        ivp.T=20.
        ivp.rhs = lambda t,u: (u-t)/(u+t)
        ivp.dt0 = 1.e-2
    elif testkey=='B1':
        ivp.u0=np.array([1.,3.])
        ivp.T=20.
        ivp.rhs = lambda t,u: np.array([2.*(u[0]-u[0]*u[1]),-(u[1]-u[0]*u[1])])
        ivp.dt0 = 1.e-2
    elif testkey=='B2':
        ivp.u0=np.array([2.,0.,1.])
        ivp.T=20.
        ivp.rhs = lambda t,u: np.array([-u[0]+u[1],u[0]-2.*u[1]+u[2],u[1]-u[2]])
        ivp.dt0 = 1.e-2
    elif testkey=='B3':
        ivp.u0=np.array([1.,0.,0.])
        ivp.T=20.
        ivp.rhs = lambda t,u: np.array([-u[0],u[0]-u[1]**2,u[1]**2])
        ivp.dt0 = 1.e-2
    elif testkey=='B4':
        ivp.u0=np.array([3.,0.,0.])
        ivp.T=20.
        ivp.rhs = _B4rhs
        ivp.dt0 = 1.e-2
    elif testkey=='B5':
        ivp.u0=np.array([0.,1.,1.])
        ivp.T=20.
        ivp.rhs = lambda t,u: np.array([u[1]*u[2],-u[0]*u[2],-0.51*u[0]*u[1]])
        ivp.dt0 = 1.e-2
    elif testkey=='C1':
        ivp.u0=np.zeros(10); ivp.u0[0]=1.
        ivp.T=20.
        e=np.ones(10); e[-1]=0.
        ivp.L_rhs = np.diag(-e)+np.diag(e[:-1],-1);
        ivp.rhs = lambda t,u: np.dot(ivp.L_rhs,u)
        ivp.dt0 = 1.e-2
    elif testkey=='C2':
        ivp.u0=np.zeros(10); ivp.u0[0]=1.
        ivp.T=20.
        e=np.arange(1,11); e[-1]=0.
        ivp.L_rhs = np.diag(-e)+np.diag(e[:-1],-1);
        ivp.rhs = lambda t,u: np.dot(ivp.L_rhs,u)
        ivp.dt0 = 1.e-2
    elif testkey=='C3':
        ivp.u0=np.zeros(10); ivp.u0[0]=1.
        ivp.T=20.
        e=np.ones(10)
        ivp.L_rhs = np.diag(-2*e)+np.diag(e[:-1],-1)+np.diag(e[:-1],1);
        ivp.rhs = lambda t,u: np.dot(ivp.L_rhs,u)
        ivp.dt0 = 1.e-2
    elif testkey=='C4':
        ivp.u0=np.zeros(51); ivp.u0[0]=1.
        ivp.T=20.
        e=np.ones(51)
        ivp.L_rhs = np.diag(-2*e)+np.diag(e[:-1],-1)+np.diag(e[:-1],1);
        ivp.rhs = lambda t,u: np.dot(ivp.L_rhs,u)
        ivp.dt0 = 1.e-2
    elif testkey=='C5':
        ivp.u0=np.zeros(30)
        ivp.u0 = np.array([3.42947415189, 3.35386959711, 1.35494901715,
            6.64145542550, 5.97156957878, 2.18231499728, 11.2630437207,
            14.6952576794, 6.27960525067, -30.1552268759, 1.65699966404,
            1.43785752721, -21.1238353380, 28.4465098142, 15.3882659679,
            -0.557160570446, 0.505696783289, 0.230578543901, -0.415570776342,
            0.365682722812, 0.169143213293, -0.325325669158, 0.189706021964,
            0.0877265322780, -0.0240476254170, -0.287659532608, -0.117219543175,
            -0.176860753121,  -0.216393453025, -0.0148647893090])
        ivp.T=20.
        ivp.rhs = _C5rhs
        ivp.dt0 = 1.e-2
    elif testkey=='D1':
        eps=0.1
        ivp.u0=np.array([1.-eps,0.,0.,np.sqrt((1.+eps)/(1.-eps))])
        ivp.T=20.
        ivp.rhs = lambda t,u: np.array([u[2],u[3],-u[0]/(u[0]**2+u[1]**2)**1.5,-u[1]/(u[0]**2+u[1]**2)**1.5])
        ivp.dt0 = 1.e-2
    elif testkey=='D2':
        eps=0.3
        ivp.u0=np.array([1.-eps,0.,0.,np.sqrt((1.+eps)/(1.-eps))])
        ivp.T=20.
        ivp.rhs = lambda t,u: np.array([u[2],u[3],-u[0]/(u[0]**2+u[1]**2)**1.5,-u[1]/(u[0]**2+u[1]**2)**1.5])
        ivp.dt0 = 1.e-2
    elif testkey=='D3':
        eps=0.5
        ivp.u0=np.array([1.-eps,0.,0.,np.sqrt((1.+eps)/(1.-eps))])
        ivp.T=20.
        ivp.rhs = lambda t,u: np.array([u[2],u[3],-u[0]/(u[0]**2+u[1]**2)**1.5,-u[1]/(u[0]**2+u[1]**2)**1.5])
        ivp.dt0 = 1.e-2
    elif testkey=='D4':
        eps=0.7
        ivp.u0=np.array([1.-eps,0.,0.,np.sqrt((1.+eps)/(1.-eps))])
        ivp.T=20.
        ivp.rhs = lambda t,u: np.array([u[2],u[3],-u[0]/(u[0]**2+u[1]**2)**1.5,-u[1]/(u[0]**2+u[1]**2)**1.5])
        ivp.dt0 = 1.e-2
    elif testkey=='D5':
        eps=0.9
        ivp.u0=np.array([1.-eps,0.,0.,np.sqrt((1.+eps)/(1.-eps))])
        ivp.T=20.
        ivp.rhs = lambda t,u: np.array([u[2],u[3],-u[0]/(u[0]**2+u[1]**2)**1.5,-u[1]/(u[0]**2+u[1]**2)**1.5])
        ivp.dt0 = 1.e-2
    elif testkey=='E1':
        ivp.u0=np.array([0.6713967071418030,0.09540051444747446])
        ivp.T=20.
        ivp.rhs = lambda t,u: np.array([u[1],-(u[1]/(t+1.) + (1.-0.25/(t+1.)**2)*u[0])])
        ivp.dt0 = 1.e-2
    elif testkey=='E2':
        ivp.u0=np.array([2.,0.])
        ivp.rhs = lambda t,u: np.array([u[1],(1.-u[0]**2)*u[1]-u[0]])
        ivp.T=20.
        ivp.dt0 = 1.e-2
    elif testkey=='E3':
        ivp.u0=np.array([0.,0.])
        ivp.rhs = lambda t,u: np.array([u[1],u[0]**3 /6. - u[0] + 2.*np.sin(2.78535*t)])
        ivp.T=20.
        ivp.dt0 = 1.e-2
    elif testkey=='E4':
        ivp.u0=np.array([30.,0.])
        ivp.rhs = lambda t,u: np.array([u[1],0.032-0.4*u[1]**2])
        ivp.T=20.
        ivp.dt0 = 1.e-2
    elif testkey=='E5':
        ivp.u0=np.array([0.,0.])
        ivp.rhs = lambda t,u: np.array([u[1],np.sqrt(1.+u[1]**2)/(25.-t)])
        ivp.T=20.
        ivp.dt0 = 1.e-2
    elif testkey=='F1':
        ivp.u0=np.array([0.,0.])
        ivp.rhs = _F1rhs
        ivp.T=20.
        ivp.dt0 = 1.e-2
    elif testkey=='F2':
        ivp.u0=np.array([110.])
        ivp.rhs = _F2rhs
        ivp.T=20.
        ivp.dt0 = 1.e-2
    elif testkey=='F3':
        ivp.u0=np.array([0.,0.])
        ivp.rhs = lambda t,u: np.array([u[1],0.01*u[1]*(1.-u[0]**2)-u[0]-np.abs(np.sin(np.pi*t))])
        ivp.T=20.
        ivp.dt0 = 1.e-2
    elif testkey=='F4':
        ivp.u0=np.array([1.])
        ivp.rhs = _F4rhs
        ivp.T=20.
        ivp.dt0 = 1.e-2
    elif testkey=='F5':
        ivp.u0=np.array([1.])
        ivp.rhs = _F5rhs
        ivp.T=20.
        ivp.dt0 = 1.e-2
    elif testkey=='SB1':
        mu=0.01212856276531231
        mudash = 1.0-mu
        ivp.u0= np.array([1.2,0.,0.,-1.049357509830319])
        ivp.T=6.192169331319639
        ivp.rhs = lambda t,u: np.array([u[2],u[3],u[0]+2.*u[3]-mudash*(u[0]+mu)/((u[0]+mu)**2+u[1]**2)**1.5-mu*(u[0]-mudash)/((u[0]-mudash)**2+u[1]**2)**1.5,u[1]-2.*u[2]-mudash*(u[1])/((u[0]+mu)**2+u[1]**2)**1.5-mu*(u[1])/((u[0]-mudash)**2+u[1]**2)**1.5])
        ivp.dt0 = 1.e-1
    elif testkey=='SB2':
        mu=0.012277471
        mudash = 1.0-mu
        ivp.u0= np.array([0.994,0,0,-2.0317326955734])
        ivp.T=11.1243403372661
        ivp.rhs = lambda t,u: np.array([u[2],u[3],u[0]+2.*u[3]-mudash*(u[0]+mu)/((u[0]+mu)**2+u[1]**2)**1.5-mu*(u[0]-mudash)/((u[0]-mudash)**2+u[1]**2)**1.5,u[1]-2.*u[2]-mudash*(u[1])/((u[0]+mu)**2+u[1]**2)**1.5-mu*(u[1])/((u[0]-mudash)**2+u[1]**2)**1.5])
        ivp.dt0 = 1.e-1
    elif testkey=='SB3':
        mu=0.012277471
        mudash = 1.0-mu
        ivp.u0 = np.array([0.994,0.,0.,-2.11389879669450])
        ivp.T=5.43679543926019
        ivp.rhs = lambda t,u: np.array([u[2],u[3],u[0]+2.*u[3]-mudash*(u[0]+mu)/((u[0]+mu)**2+u[1]**2)**1.5-mu*(u[0]-mudash)/((u[0]-mudash)**2+u[1]**2)**1.5,u[1]-2.*u[2]-mudash*(u[1])/((u[0]+mu)**2+u[1]**2)**1.5-mu*(u[1])/((u[0]-mudash)**2+u[1]**2)**1.5])
        ivp.dt0 = 1.e-1

    else: raise Exception('Unknown Detest problem')
    ivp.name=testkey
    ivp.description='Problem '+testkey+' of the non-stiff DETEST suite.'
    return ivp

def _F1rhs(t,u):
    a=0.1
    du=np.zeros(2)
    du[0]=u[1]
    if np.mod(np.floor(t),2):     #Odd
      du[1] = 2.*a*u[1] - (np.pi**2+a**2)*u[0] - 1
    else:
      du[1] = 2.*a*u[1] - (np.pi**2+a**2)*u[0] + 1
    return du

def _F2rhs(t,u):
    du=np.zeros(1)
    if np.mod(np.floor(t),2):     #Odd
      du[0] = 55.-u[0]/2.
    else:
      du[0] = 55.-3.*u[0]/2.
    return du

def _F4rhs(t,u):
    du=np.zeros(1)
    if t<=10.:
      du[0] = -2./21 - 120.*(t-5.)/(1.+4.*(t-5.)**2)
    else:
      du[0] = -2.*u[0]
    return du

def _F5rhs(t,u):
    du=np.zeros(1)
    c=np.sum([i**(4./3) for i in range(1,20)])
    du[0] = 4./3/c * np.sum([(t-i+0j)**(4./3) for i in range(1,20)]).real*u[0]
    return du

def _B4rhs(t,u):
    du=np.zeros(3)
    du[0]=-u[1] - (u[0]*u[2])/np.sqrt(u[0]**2+u[1]**2)
    du[1]= u[0] - (u[1]*u[2])/np.sqrt(u[0]**2+u[1]**2)
    du[2]=        (u[0]     )/np.sqrt(u[0]**2+u[1]**2)
    return du
def _C5rhs(t,u):
    r = np.zeros(5); u1 = np.zeros(15); u2 = np.zeros(15); y1 = np.zeros((3,5)); y2 = np.zeros((3,5)); d = np.zeros((5,5)); du=np.zeros(30); ynew1 = np.zeros((3,5)); ynew2 = np.zeros((3,5));
    k2 = 2.95912208286; m0 = 1.00000597682;
    m = np.array([.000954786104043,.000285583733151,.0000437273164546,.0000517759138449,.00000277777777778])
    k=0
    for j in range(0,5):
        for i in range(0,3):
            y1[i][j] = u[k]
            y2[i][j] = u[k+15]
            k+=1

    for j in range(0,5):
        r[j] = np.sqrt(y1[0][j]**2+y1[1][j]**2+y1[2][j]**2)

    for k in range(0,5):
        for j in range(0,5):
            d[k][j] = np.sqrt((y1[0][k]-y1[0][j])**2+(y1[1][k]-y1[1][j])**2+(y1[2][k]-y1[2][j])**2)

    for j  in range(0,5):
        for i in range(0,3):
            term=0.;
            for k in range(0,5):
                if (k != j):
                    term = term+m[k]*((y1[i][k]-y1[i][j])/d[j][k]**3 - y1[i][k]/r[k]**3)
            ynew1[i][j] = y2[i][j]
            ynew2[i][j] = k2*(-(m0+m[j])*y1[i][j]/r[j]**3+term)
    k=0
    for j in range(0,5):
        for i in range(0,3):
            du[k] = ynew1[i][j]
            du[k+15] = ynew2[i][j]
            k+=1

    return du

def detest_suite():
    """The entire non-stiff DETEST suite of problems."""
    detestkeys=['A1','A2','A3','A4','A5','B1','B2','B3','B4','B5','C1','C2','C3','C4','C5','D1','D2','D3','D4','D5','E1','E2','E3','E4','E5','F1','F2','F3','F4','F5']
    return [detest(dtkey) for dtkey in detestkeys]

def detest_suite_plus():
    """The entire non-stiff DETEST suite of problems plus the problems from Shampine-Baca paper. More problems can be added"""
    detestkeys=['A1','A2','A3','A4','A5','B1','B2','B3','B4','B5','C1','C2','C3','C4','C5','D1','D2','D3','D4','D5','E1','E2','E3','E4','E5','F1','F2','F3','F4','F5','SB1','SB2','SB3']
    return [detest(dtkey) for dtkey in detestkeys]

def detest_suite_minus():
    """The entire non-stiff DETEST suite of problems except the F (discontinous) problems."""
    detestkeys=['A1','A2','A3','A4','A5','B1','B2','B3','B4','B5','C1','C2','C3','C4','C5','D1','D2','D3','D4','D5','E1','E2','E3','E4','E5']
    return [detest(dtkey) for dtkey in detestkeys]

def detest_stiff_suite():
    """The entire non-stiff DETEST suite of problems."""
    detestkeys=['A1', 'A2', 'A3', 'A4']
    return [detest_stiff(dtkey) for dtkey in detestkeys]


def detest_stiff(testkey):
    """The stiff DETEST suite.  Only a few problems have been implemented."""
    ivp=IVP()
    if testkey=='A1':
        ivp.u0=np.array([1.,1.,1.,1.])
        ivp.T=20.
        ivp.rhs = lambda t,u: np.array([-0.5*u[0], -u[1], -100.*u[2], -90.*u[3]])
        ivp.dt0 = 1.e-2
    elif testkey=='A2':
        ivp.u0=np.array([0.,0.,0.,0.,0.,0.,0.,0.,0.])
        ivp.T=120.
        ivp.rhs = _A2rhs_stiff
        ivp.dt0 = 5.e-4
    elif testkey=='A3':
        ivp.u0=np.array([1.,1.,1.,1.])
        ivp.T=20.
        ivp.rhs = _A3rhs_stiff
        ivp.dt0 = 1.e-5
    elif testkey=='A4':
        ivp.u0=np.zeros(10)
        ivp.T=1.
        ivp.rhs = _A4rhs_stiff
        ivp.dt0 = 1.e-5
    else:
        print('Unknown Detest problem; returning empty IVP')
    return ivp

def _A2rhs_stiff(t,u):
    du=np.zeros(9)
    du[0]=-1800.*u[0] + 900.*u[1]
    for i in range(1,8):
        du[i]=u[i-1]-2.*u[i]+u[i+1]
    du[8]=1000.*u[7]-2000.*u[8]+1000.
    return du

def _A3rhs_stiff(t,u):
    du=np.zeros(4)
    du[0]=-1.e4*u[0] + 100.*u[1] - 10.*u[2] + u[3]
    du[1]=-1.e3*u[1] + 10.*u[2] - 10.*u[3]
    du[2]=-u[2]+10.*u[3]
    du[3]=-0.1*u[3]
    return du

def _A4rhs_stiff(t,u):
    du=np.zeros(10)
    for i in range(10):
        du[i]=-(i+1)**5. * u[i]
    return du


if __name__ == "__main__":
    import doctest
    doctest.testmod()



"""
Class for Downwind Runge-Kutta methods, and various functions related to them.

**Examples**::

    >>> from nodepy import downwind_runge_kutta_method as dwrk

* Load a method::

    >>> dw2 = dwrk.opt_dwrk(8)
    >>> print(dw2) # doctest: +NORMALIZE_WHITESPACE
    downwind Runge-Kutta Method
    <BLANKLINE>
     0.750 |  2.875         |         2.125
     0.875 |  3.000         |         2.125
    _______|_________________________________
           |  3.000  0.125 |          2.125

* Check some of its properties::

    >>> dw2.order()
    2
    >>> dw2.absolute_monotonicity_radius()
    7.999999999992724


REFERENCES:
    [higueras2005,Ketcheson2010]_
"""
from __future__ import division

from __future__ import absolute_import
import numpy as np
import nodepy.snp as snp
import nodepy.runge_kutta_method as rk
from nodepy.general_linear_method import GeneralLinearMethod
from six.moves import range

#=====================================================
class DownwindRungeKuttaMethod(GeneralLinearMethod):
#=====================================================
    """ General class for Downwind Runge-Kutta Methods """
    def __init__(self,A=None,At=None,b=None,bt=None,alpha=None,beta=None,
            alphat=None,betat=None,
            name='downwind Runge-Kutta Method',description=''):
        r"""
            Initialize a downwind Runge-Kutta method.  The representation
            uses the form and notation of [Ketcheson2010]_.

            \\begin{align*}
            y^n_i = & u^{n-1} + \\Delta t \\sum_{j=1}^{s}
            (a_{ij} f(y_j^{n-1}) + \\tilde{a}_{ij} \\tilde{f}(y_j^n)) & (1\\le j \\le s) \\\\
            \\end{align*}

        """
        A,b,alpha,beta=snp.normalize(A,b,alpha,beta)

        butchform=[x is not None for x in [A,At,b,bt]]
        SOform=[x is not None for x in [alpha,alphat,beta,betat]]
        if not ( ( all(butchform) and not (True in SOform) ) or
                    ( (not (True in butchform)) and all(SOform) ) ):
            raise Exception("""To initialize a Runge-Kutta method,
                you must provide either Butcher arrays or Shu-Osher arrays,
                but not both.""")
        if A is not None: #Initialize with Butcher arrays
            # Check that number of stages is consistent
            m=np.size(A,0) # Number of stages
            if m>1:
                if not np.all([np.size(A,1),np.size(b)]==[m,m]):
                   raise Exception('Inconsistent dimensions of Butcher arrays')
            else:
                if not np.size(b)==1:
                    raise Exception('Inconsistent dimensions of Butcher arrays')
        elif alpha is not None: #Initialize with Shu-Osher arrays
            A,At,b,bt=downwind_shu_osher_to_butcher(alpha,alphat,beta,betat)
        # Set Butcher arrays
        if len(np.shape(A))==2:
          self.A=A
          self.At=At
        else:
          self.A =snp.array([A ]) #Fix for 1-stage methods
          self.At=snp.array([At])
        self.b=b;
        self.bt=bt
        self.c=np.sum(self.A,1)-np.sum(self.At,1)
        self.name=name
        self.info=description
        self.underlying_method=rk.RungeKuttaMethod(self.A-self.At,self.b-self.bt)

    def __repr__(self):
        """
        Pretty-prints the Butcher array in the form:
          |   |
        c | A | At
        ___________
          | b | bt
        """
        from nodepy.utils import shortstring
        c = [shortstring(ci) for ci in self.c]
        clenmax = max([len(ci) for ci in c])
        A = [shortstring(ai) for ai in self.A.reshape(-1)]
        alenmax = max([len(ai) for ai in A])
        b = [shortstring(bi) for bi in self.b]
        blenmax = max([len(bi) for bi in b])
        At = [shortstring(ai) for ai in self.At.reshape(-1)]
        atlenmax = max([len(ai) for ai in At])
        bt = [shortstring(bi) for bi in self.bt]
        btlenmax = max([len(bi) for bi in bt])
        colmax=max(alenmax,blenmax)
        colmax2 = max(atlenmax, btlenmax)
        colmax = max(colmax,colmax2)


        s=self.name+'\n'+self.info+'\n'
        for i in range(len(self)):
            s+=c[i]+' '*(clenmax-len(c[i])+1)+'| '
            for j in range(len(self)):
                ss=shortstring(self.A[i,j])
                s+=ss.ljust(colmax+1)
            s+=' | '
            for j in range(len(self)):
                ss=shortstring(self.At[i,j])
                s+=ss.ljust(colmax+1)
            s+='\n'
        s+='_'*(clenmax+1)+'|'+('_______'*2*len(self))+('_____')+'\n'
        s+= ' '*(clenmax+1)+'|'
        for j in range(len(self)):
            s+=' '*(colmax-len(b[j])+1)+b[j]
        s+=' | '
        for j in range(len(self)):
            s+=' '*(colmax-len(bt[j])+1)+bt[j]
        return s

    def __len__(self):
        """
            The length of the method is the number of stages.
        """
        return np.size(self.A,0)

    def order(self,tol=1.e-13):
        r"""
            Return the order of a Downwind Runge-Kutta method.
        """
        return self.underlying_method.order(tol)

    def absolute_monotonicity_radius(self,acc=1.e-10,rmax=200,
                    tol=3.e-16):
        r"""
            Returns the radius of absolute monotonicity
            of a Runge-Kutta method.
        """
        from nodepy.utils import bisect

        r=bisect(0,rmax,acc,tol,self.is_absolutely_monotonic)
        return r

    def is_absolutely_monotonic(self,r,tol):
        r""" Returns 1 if the downwind Runge-Kutta method is
            absolutely monotonic at $z=-r$.

            The method is absolutely monotonic if $(I+rK+rKt)^{-1}$ exists
            and
            $$(I+rK+rKt)^{-1}K \\ge 0$$
            $$(I+rK+rKt)^{-1}Kt \\ge 0$$
            $$(I+rK+rKt)^{-1} e_m \\ge 0$$

            where $e_m$ is the m-by-1 vector of ones and
                  K=[ K  0
                     b^T 0].

            The inequalities are interpreted componentwise.

        """
        m=len(self)
        K  =np.hstack([np.vstack([self.A  ,self.b ]),np.zeros([m+1,1])]).astype(np.float64)
        Kt =np.hstack([np.vstack([self.At ,self.bt]),np.zeros([m+1,1])]).astype(np.float64)
        X=np.eye(len(self)+1) + r*(K+Kt)
        beta =r*np.linalg.solve(X, K)
        betat=r*np.linalg.solve(X, Kt)
        ech=np.linalg.solve(X,np.ones(m+1))
        if min(beta.min(),betat.min(),ech.min())<-tol:
            return 0
        else:
            return 1
        # Need an exception here if rhi==rmax

#================================================================

def opt_dwrk(r):
    #a11=(r**2-2*r-2)/(2.*r)
    at12=(r**2-4*r+2)/(2*r)
    #a21=(r-2.)/2.
    alpha=np.array([[2/r,0],[1,0],[0,1]])
    alphat=np.array([[0,(r**2-4*r+2)/(r**2-2*r)],[0,0],[0,0]])
    beta=alpha/r
    betat=alphat/r
    return DownwindRungeKuttaMethod(alpha=alpha,beta=beta,alphat=alphat,betat=betat)


def downwind_shu_osher_to_butcher(alpha,alphat,beta,betat):
    r""" Accepts a Shu-Osher representation of a downwind Runge-Kutta
        method and returns the Butcher coefficients

        \\begin{align*}
        A  = & (I-\\alpha_0-\\alphat_0)^{-1} \\beta_0 \\\\
        At = & (I-\\alpha_0-\\alphat_0)^{-1} \\betat_0 \\\\
        b = & \\beta_1 + (\\alpha_1 + \\alphat_1) * A
        \\end{align*}

        **References**:
             #. [gottlieb2009]_
    """
    m=np.size(alpha,1)
    if not np.all([np.size(alpha,0),np.size(beta,0),
                    np.size(beta,1)]==[m+1,m+1,m]):
        raise Exception('Inconsistent dimensions of Shu-Osher arrays')
    X=snp.eye(m)-alpha[0:m,:]-alphat[0:m,:]
    A =snp.solve(X, beta[0:m,:])
    At=snp.solve(X,betat[0:m,:])
    b=beta[m,:]+np.dot(alpha[m,:]+alphat[m,:],A)
    bt=betat[m,:]+np.dot(alpha[m,:]+alphat[m,:],At)
    return A,At,b,bt


if __name__ == "__main__":
    import doctest
    doctest.testmod()



"""
Functions for running convergence and performance tests.

**Examples**::

    >>> from nodepy import rk, convergence, ivp

    # Load some methods:
    >>> rk4=rk.loadRKM('RK44')
    >>> SSP2=rk.loadRKM('SSP22')
    >>> SSP104=rk.loadRKM('SSP104')

    # Define an initial value problem:
    >>> myivp=ivp.load_ivp('test')

    >>> work, error = convergence.ctest([rk4,SSP2,SSP104],myivp)
"""
from __future__ import print_function

from __future__ import absolute_import
import numpy as np
import nodepy.runge_kutta_method as rk

def ctest(methods,ivp,grids=[20,40,80,160,320,640],verbosity=0,parallel=False):
    """
        Runs a convergence test, integrating a single initial value problem
        using a sequence of fixed step sizes and a set of methods.
        Creates a plot of the resulting errors versus step size for each method.

        **Inputs**:
            - methods -- a list of ODEsolver instances
            - ivp     -- an IVP instance
            - grids   -- a list of grid sizes for integration.
                      optional; defaults to [20,40,80,160,320,640]
            - parallel -- to exploit possible parallelization (optional)

        **Example**::

            >>> import nodepy.runge_kutta_method as rk
            >>> from nodepy.ivp import load_ivp
            >>> rk44=rk.loadRKM('RK44')
            >>> myivp=load_ivp('nlsin')
            >>> work, err=ctest(rk44,myivp)

        TODO:
            - Option to plot versus f-evals or dt

    """
    import matplotlib.pyplot as plt
    plt.clf();
    # In case just one method is passed in (and not as a list):
    if not isinstance(methods,list): methods=[methods]
    err=[]
    try:
        exsol = ivp.exact(ivp.T)
    except:
        bs5=rk.loadRKM('BS5')
        bigN=grids[-1]*4
        if verbosity>0: print('solving on fine grid with {} points'.format(bigN))
        t,u=bs5(ivp,N=bigN)
        if verbosity>0: print('done')
        exsol = u[-1] + 0.
    for method in methods:
        if verbosity>0: print("solving with {}".format(method.name))
        err0=[]
        for i,N in enumerate(grids):
            t,u=method(ivp,N=N)
            err0.append(np.linalg.norm(u[-1]-exsol))
        err.append(err0)
        work=np.array([grid*len(method) for grid in grids])
        if parallel:
            speedup = len(method)/float(method.num_seq_dep_stages())
            work = work/speedup
        plt.loglog(work,err0,label=method.name,linewidth=3)
    plt.xlabel('Function evaluations')
    plt.ylabel('Error at $t_{final}$')
    plt.legend(loc='best')
    plt.draw()
    return work, err


def ptest(methods,ivps,tols=[1.e-1,1.e-2,1.e-4,1.e-6],verbosity=0,parallel=False,
          controllertype='P'):
    """
        Runs a performance test, integrating a set of problems with a set
        of methods using a sequence of error tolerances.  Creates a plot
        of the error achieved versus the amount of work done (number of
        function evaluations) for each method.

        **Input**:
            * methods -- a list of ODEsolver instances
                      Note that all methods must have error estimators.
            * ivps    -- a list of IVP instances
            * tols    -- a specified list of error tolerances (optional)
            * parallel -- to exploit possible parallelization (optional)

        **Example**::

            >>> import nodepy.runge_kutta_method as rk
            >>> from nodepy.ivp import load_ivp
            >>> bs5=rk.loadRKM('BS5')
            >>> myivp=load_ivp('nlsin')
            >>> work,err=ptest(bs5,myivp)

    """
    import matplotlib.pyplot as plt
    plt.clf(); plt.draw();
    # In case just one method is passed in (and not as a list):
    if not isinstance(methods,list): methods=[methods]
    if not isinstance(ivps,list): ivps=[ivps]
    err=np.ones([len(methods),len(tols)])
    work=np.zeros([len(methods),len(tols)])
    for ivp in ivps:
        if verbosity>0: print("solving problem {}".format(ivp))
        try:
            exsol = ivp.exact(ivp.T)
        except:
            bs5=rk.loadRKM('BS5')   #Use Bogacki-Shampine RK for fine solution
            lowtol=min(tols)/100.
            if verbosity>0: print('solving for "exact" solution with tol= {}'.format(lowtol))
            t,u=bs5(ivp,errtol=lowtol,dt=ivp.dt0)
            if verbosity>0: print('done')
            exsol = u[-1] + 0.
        for imeth,method in enumerate(methods):
            if verbosity>0: print('Solving with method {}'.format(method.name))
            if parallel:
                speedup = len(method)/float(method.num_seq_dep_stages())
            else:
                speedup = 1.
            workperstep = len(method)-method.is_FSAL()
            for jtol,tol in enumerate(tols):
                t,u,rej,dt,errhist=method(ivp,errtol=tol,dt=ivp.dt0,diagnostics=True,controllertype=controllertype)
                if verbosity>1: print('{} rejected steps'.format(rej))
                err[imeth,jtol]*= np.max(np.abs(u[-1]-exsol))
                #FSAL methods save on accepted steps, but not on rejected:
                work[imeth,jtol]+= (len(t)*workperstep+rej*len(method))/speedup
    for imeth,method in enumerate(methods):
        for jtol,tol in enumerate(tols):
            err[imeth,jtol]=err[imeth,jtol]**(1./len(ivps))
    for imeth,method in enumerate(methods):
        plt.semilogy(work[imeth,:],err[imeth,:],label=method.name,linewidth=3)
    plt.xlabel('Function evaluations')
    plt.ylabel('Error at $t_{final}$')
    plt.legend(loc='best')
    plt.draw()
    return work,err



if __name__ == "__main__":
    import doctest
    doctest.testmod()



r"""
Load some high-order RK pairs.

**Examples**::

    >>> from nodepy import loadmethod
    >>> myrk = loadmethod.load_rk_from_file()
    >>> myrk.order()
    12
    >>> myrk.name
    "Feagin's 12(10) pair"

"""
from __future__ import absolute_import
import os
import numpy as np

from nodepy import rk

path = os.path.dirname(os.path.realpath(__file__))
path = os.path.join(path,'method_coefficients')

def method_name_and_stages(file_name):
    if ((file_name=='rk108.txt')):
        name=r"Feagin's 10(8) pair"
        stages=20
    if ((file_name=='rk108curtis.txt')):
        name=r"Curtis' 10(8) pair"
        stages=21
    if ((file_name=='rk1210.txt')):
        name=r"Feagin's 12(10) pair"
        stages=25
    if ((file_name=='rk129hiroshi.txt')):
        name=r"Hiroshi's 12(9) pair"
        stages=29
    if ((file_name=='rk1412.txt')):
        name=r"Feagin's 14(12) pair"
        stages=35
    return name, stages

def load_rk_from_file(file_name='rk1210.txt', load_pair=True):
    name, stages = method_name_and_stages(file_name)
    f = open(os.path.join(path,file_name))
    b = np.zeros((stages))
    bhat = np.zeros((stages))
    A = np.zeros((stages,stages))
    c = np.zeros((stages))

    for line in f.readlines():
        words = line.split()
        if words == []:
            continue

        if words[-1]=='b[k]':
            current_array = 'b'
        elif words[-1]=='c[k]':
            current_array = 'c'
        elif words[-1]=='bhat[k]':
            current_array = 'bhat'
        elif words[-1]=='A[k,j]':
            current_array = 'A'

        try:
            k = int(words[0])
            if current_array == 'b':
                b[k] = float(words[1])
            elif current_array =='bhat':
                bhat[k] = float(words[1])
            elif current_array == 'A':
                j = int(words[1])
                A[k,j] = float(words[2])
        except ValueError:
            continue

    if load_pair:
        return rk.ExplicitRungeKuttaPair(A=A, b=b, bhat=bhat, name=name)
    else:
        return rk.ExplicitRungeKuttaMethod(A=A, b=b, name=name)



from __future__ import print_function

from __future__ import absolute_import
import numpy as np
from six.moves import range
import sympy

def E_polynomial(p, q=None):
    r"""
        Compute the E-polynomial discussed in Hairer & Wanner.

        For R(z) = P(z)/Q(z), the E-polynomial is

        E(y) = Q(iy) Q(-iy) - P(iy) P(-iy).

        The method is A-stable if the E-polynomial is non-negative
        for all real y.
    """
    if q is None:
        q = np.poly1d([1])

    from copy import deepcopy
    piy  = deepcopy(p)
    pmiy = deepcopy(p)
    qiy  = deepcopy(q)
    qmiy = deepcopy(q)

    for i in range(1,len(p.c)+1):
        piy.coeffs[-i] = (sympy.I)**(i-1)*piy.coeffs[-i]
        pmiy.coeffs[-i] = (-sympy.I)**(i-1)*pmiy.coeffs[-i]

    for i in range(1,len(q.c)+1):
        qmiy.coeffs[-i] = (-sympy.I)**(i-1)*qmiy.coeffs[-i]
        qiy.coeffs[-i] = (sympy.I)**(i-1)*qiy.coeffs[-i]

    Epoly = qiy*qmiy - piy*pmiy
    return Epoly

def imaginary_stability_interval(p,q=None,eps=1.e-14):
    r"""
        Length of imaginary axis half-interval contained in the
        method's region of absolute stability.

        **Examples**::

            >>> from nodepy import rk
            >>> rk4 = rk.loadRKM('RK44')
            >>> rk4.imaginary_stability_interval() #doctest: +ELLIPSIS
            2.8284271247...
    """
    if q is None:
        q = np.poly1d([1.])

    c = p.c[::-1].copy()
    c[1::2] = 0      # Zero the odd coefficients to get real part
    c[::2][1::2] = -1*c[::2][1::2]  # Negate coefficients with even powers of i
    p1 = np.poly1d(c[::-1])

    c = p.c[::-1].copy()
    c[::2] = 0      # Zero the even coefficients to get imaginary part
    c[1::2][1::2] = -1*c[1::2][1::2]  # Negate coefficients with even powers of i
    p2 = np.poly1d(c[::-1])

    c = q.c[::-1].copy()
    c[1::2] = 0      # Zero the odd coefficients to get real part
    c[::2][1::2] = -1*c[::2][1::2]  # Negate coefficients with even powers of i
    q1 = np.poly1d(c[::-1])

    c = q.c[::-1].copy()
    c[::2] = 0      # Zero the even coefficients to get imaginary part
    c[1::2][1::2] = -1*c[1::2][1::2]  # Negate coefficients with even powers of i
    q2 = np.poly1d(c[::-1])

    ppq = p1**2 + p2**2 + q1**2 + q2**2
    pmq = p1**2 + p2**2 - q1**2 - q2**2

    ppq_roots = np.array([x.real for x in ppq.r if abs(x.imag)<eps and x.real>0])
    pmq_roots = np.array([x.real for x in pmq.r if abs(x.imag)<eps and x.real>0])

    if len(pmq_roots)>0: pmqr = np.min(pmq_roots)
    else: pmqr = np.inf
    if len(ppq_roots)>0: ppqr = np.min(ppq_roots)
    else: ppqr = np.inf

    mr = min(pmqr,ppqr)

    if mr == np.inf:
        # Stability region boundary does not touch/cross the imaginary axis
        # If it's explicit, it must be unstable for all imaginary values
        if len(q.coeffs==1): return 0
        z = 1j/2.
    else:
        z = mr*1j/2.

    val = 1
    while val==1:
        # Check whether it is stable between 0 and mr
        # This could be wrong if the stability boundary is tangent to the imaginary axis at mr
        val = np.abs(p(z)/q(z))
        if val<1:
            return mr
        elif val==1:
            z = z/2.
        else:
            return 0
        if np.abs(z)<1.e-10:
            print("Warning: unable to determine exact imaginary stability interval")
            return mr


def real_stability_interval(p,q=None,eps=1.e-12):
    r"""
        Length of negative real axis interval contained in the
        method's region of absolute stability.

        **Examples**::

            >>> from nodepy import rk
            >>> rk4 = rk.loadRKM('RK44')
            >>> I = rk4.real_stability_interval()
            >>> print("{:.10f}".format(I))
            2.7852935634
            >>> rkc = rk.RKC1(2)
            >>> rkc.real_stability_interval()
            8.0
    """
    if q is None: q = np.poly1d([1.])

    # Find points where p = +/- q
    pmq = p-q
    ppq = p+q
    pmq_roots = np.array([-x.real for x in pmq.r if abs(x.imag)<eps and x.real<0])
    ppq_roots = np.array([-x.real for x in ppq.r if abs(x.imag)<eps and x.real<0])
    roots = np.hstack((pmq_roots,ppq_roots))
    roots = np.unique(roots)

    z = -roots[0]/2.
    if np.abs(p(z)/q(z))>1.+eps:
        return 0.

    for i in range(len(roots)-1):
        z = -(roots[i]+roots[i+1])/2
        if np.abs(p(z)/q(z))>1.+eps:
            return roots[i]
    z = -roots[-1]*2
    if np.abs(p(z)/q(z))>1.+eps:
        return roots[-1]
    return np.inf


def plot_stability_region(p,q,N=200,color='r',filled=True,bounds=None,
                          plotroots=False,alpha=1.,scalefac=1.,fignum=None):
    r"""
        Plot the region of absolute stability of a rational function; i.e. the set

        `\{ z \in C : |\phi (z)|\le 1 \}`

        Unless specified explicitly, the plot bounds are determined automatically, attempting to
        include the entire region.  A check is performed beforehand for
        methods with unbounded stability regions.
        Note that this function is not responsible for actually drawing the
        figure to the screen (or saving it to file).

        **Inputs**:
            - p, q  -- Numerator and denominator of the stability function
            - N       -- Number of gridpoints to use in each direction
            - color   -- color to use for this plot
            - filled  -- if true, stability region is filled in (solid); otherwise it is outlined
            - plotroots -- if True, plot the roots and poles of the function
            - alpha -- transparency of contour plot
            - scalefac -- factor by which to scale region (often used to normalize for stage number)
            - fignum -- number of existing figure to use for plot

    """
    import matplotlib.pyplot as plt

    # Convert coefficients to floats for speed
    if p.coeffs.dtype=='object':
        p = np.poly1d([float(c) for c in p.coeffs])
    if q.coeffs.dtype=='object':
        q = np.poly1d([float(c) for c in q.coeffs])

    if bounds is None:
        from nodepy.utils import find_plot_bounds
        # Check if the stability region is bounded or not
        m,n = p.order,q.order
        if (m < n) or ((m == n) and (abs(p[m])<abs(q[n]))):
            print('The stability region is unbounded')
            if m > 0:
                bounds = (-10*m,m,-5*m,5*m)
            else:
                bounds = (-10,1,-5,5)
        else:
            stable = lambda z : np.abs(p(z)/q(z))<=1.0
            bounds = find_plot_bounds(stable,guess=(-10,1,-5,5))
            if np.min(np.abs(np.array(bounds)))<1.e-14:
                print('No stable region found; is this method zero-stable?')

        if (m == n) and (abs(p[m])==abs(q[n])):
            print('The stability region may be unbounded')

    # Evaluate the stability function over a grid
    x=np.linspace(bounds[0],bounds[1],N)
    y=np.linspace(bounds[2],bounds[3],N)
    X=np.tile(x,(N,1))
    Y=np.tile(y[:,np.newaxis],(1,N))
    Z=X+Y*1j
    R=np.abs(p(Z*scalefac)/q(Z*scalefac))

    # Plot
    h = plt.figure(fignum)
    if filled:
        plt.contourf(X,Y,R,[0,1],colors=color,alpha=alpha)
    else:
        plt.contour(X,Y,R,[0,1],colors=color,alpha=alpha,linewidths=3)
    if plotroots: plt.plot(np.real(p.r),np.imag(p.r),'ok')
    if len(q)>1: plt.plot(np.real(q.r),np.imag(q.r),'xk')
    plt.plot([0,0],[bounds[2],bounds[3]],'--k',linewidth=2)
    plt.plot([bounds[0],bounds[1]],[0,0],'--k',linewidth=2)
    plt.axis('Image')
    return h


def plot_order_star(p,q,N=200,bounds=(-5,5,-5,5), plotroots=False,
                color=('w','b'),fignum=None):
    r""" Plot the order star of a rational function
        i.e. the set

        $$ \{ z \in C : |R(z)/exp(z)|\le 1 \} $$

        where $R(z)=p(z)/q(z)$ is the stability function of the method.

        **Input**: (all optional)
            - N       -- Number of gridpoints to use in each direction
            - bounds  -- limits of plotting region
            - color   -- color to use for this plot
    """
    # Convert coefficients to floats for speed
    if p.coeffs.dtype=='object':
        p = np.poly1d([float(c) for c in p.coeffs])
    if q.coeffs.dtype=='object':
        q = np.poly1d([float(c) for c in q.coeffs])


    import matplotlib.pyplot as plt
    x=np.linspace(bounds[0],bounds[1],N)
    y=np.linspace(bounds[2],bounds[3],N)
    X=np.tile(x,(N,1))
    Y=np.tile(y[:,np.newaxis],(1,N))
    Z=X+Y*1j
    R=np.abs(p(Z)/q(Z)/np.exp(Z))

    h = plt.figure(fignum)
    plt.contourf(X,Y,R,[0,1,1.e299],colors=color)
    if plotroots: plt.plot(np.real(p.r),np.imag(p.r),'ok')
    plt.plot([0,0],[bounds[2],bounds[3]],'--k')
    plt.plot([bounds[0],bounds[1]],[0,0],'--k')
    plt.axis('Image')
    return h


def pade_exp(k,j):
    """
    Return the Pade approximation to the exponential function
    with numerator of degree k and denominator of degree j.

    **Example**::

        >>> from nodepy import stability_function
        >>> p, q = stability_function.pade_exp(2,3)
        >>> p
        poly1d([1/20, 2/5, 1], dtype=object)
        >>> q
        poly1d([-1/60, 3/20, -3/5, 1], dtype=object)

    """
    from sympy import Rational
    one = Rational(1)
    Pcoeffs=[one]
    Qcoeffs=[one]
    for n in range(1,k+1):
        newcoeff=Pcoeffs[0]*(k-n+one)/(j+k-n+one)/n
        Pcoeffs=[newcoeff]+Pcoeffs
    P=np.poly1d(Pcoeffs)
    for n in range(1,j+1):
        newcoeff=-one*Qcoeffs[0]*(j-n+one)/(j+k-n+one)/n
        Qcoeffs=[newcoeff]+Qcoeffs
    Q=np.poly1d(Qcoeffs)
    return P,Q

def taylor(p):
    r"""
        Return the Taylor polynomial of the exponential, up to order p.

        **Example**::

            >>> from nodepy import stability_function
            >>> stability_function.taylor(3)
            poly1d([1/6, 1/2, 1, 1], dtype=object)


poly1d([-1/60, 3/20, -3/5, 1], dtype=object)
    """
    from sympy import factorial

    coeffs = np.array( [1/factorial(k) for k in range(p+1) ] )

    return np.poly1d(coeffs[::-1])

if __name__ == "__main__":
    import doctest
    doctest.testmod()

